import torch
import torch.nn as nn
import torch.nn.functional as F

from .utils import PlaceHolder
from ...nn import MLP, AttentionWithNodeMask
from ...nn.embedder import TimestepEmbedder, CategoricalEmbedder, ClusterContinuousEmbedder

def modulate(x, shift, scale):
    return x * (1 + scale.unsqueeze(1)) + shift.unsqueeze(1)

class Transformer(nn.Module):
    def __init__(
        self,
        max_node,
        hidden_size=384,
        num_layer=12,
        num_head=16,
        mlp_ratio=4.0,
        dropout=0.,
        drop_condition=0.1,
        input_dim_X=118,
        input_dim_E=5,
        input_dim_y=None,
        task_type=[], # a list of 'regression' or 'classification'
    ):
        super().__init__()
        self.input_dim_y = input_dim_y
        self.x_embedder = nn.Linear(input_dim_X + max_node * input_dim_E, hidden_size, bias=False)
        self.t_embedder = TimestepEmbedder(hidden_size)

        if input_dim_y > 0 and len(task_type) > 0:
            self.y_embedder_list = torch.nn.ModuleList()
            for i in range(input_dim_y):
                if task_type[i] == 'regression':
                    self.y_embedder_list.append(ClusterContinuousEmbedder(1, hidden_size, drop_condition))
                else:
                    self.y_embedder_list.append(CategoricalEmbedder(2, hidden_size, drop_condition))
        else:
            self.y_embedder_list = None

        self.blocks = nn.ModuleList(
            [
                AttentionBlock(hidden_size, num_head, mlp_ratio=mlp_ratio, dropout=dropout)
                for _ in range(num_layer)
            ]
        )

        self.final_layer = FinalLayer(
            max_node=max_node,
            hidden_size=hidden_size,
            atom_type=input_dim_X,
            bond_type=input_dim_E,
            mlp_ratio=mlp_ratio,
            num_head=num_head,
        )

        self.initialize_parameters()

    def initialize_parameters(self, seed=None):
        if seed is not None:
            torch.manual_seed(seed)

        # Initialize transformer layers:
        def _basic_init(module):
            if isinstance(module, nn.Linear):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    nn.init.constant_(module.bias, 0)

        def _constant_init(module, i):
            if isinstance(module, nn.Linear):
                nn.init.constant_(module.weight, i)
                if module.bias is not None:
                    nn.init.constant_(module.bias, i)

        self.apply(_basic_init)

        for block in self.blocks :
            _constant_init(block.adaLN_modulation[0], 0)
        _constant_init(self.final_layer.adaLN_modulation[0], 0)

    def forward(self, noisy_data, unconditioned):
        X, E = noisy_data['X_t'].float(), noisy_data['E_t'].float()
        X_in, E_in = X, E
        node_mask, t =  noisy_data['node_mask'], noisy_data['t']
        y = noisy_data['y_t'] if self.y_embedder_list is not None else None
        assert (y is None) == (self.y_embedder_list is None), "y and y_embedder_list must both be None or both be not None"

        if y is not None:
            force_drop_id = torch.zeros_like(y.sum(-1))
            force_drop_id[torch.isnan(y.sum(-1))] = 1
            if unconditioned:
                force_drop_id = torch.ones_like(y[:, 0])
        
        bs, n, _ = X.size()
        h = torch.cat([X, E.reshape(bs, n, -1)], dim=-1)
        h = self.x_embedder(h)
        c = self.t_embedder(t)
        if self.y_embedder_list is not None:
            for i in range(self.input_dim_y):
                c = c + self.y_embedder_list[i](y[:, i:i+1], self.training, force_drop_id)
        
        for i, block in enumerate(self.blocks):
            h = block(h, c, node_mask)

        X, E, y = self.final_layer(h, X_in, E_in, c, t, node_mask)
        return PlaceHolder(X=X, E=E, y=y).mask(node_mask)
    
    def compute_loss(self, noisy_data, true_X, true_E, lw_X, lw_E, unconditioned=False):
        pred = self.forward(noisy_data, unconditioned=unconditioned)
        
        # Reshape predictions and targets
        true_X = torch.reshape(true_X, (-1, true_X.size(-1)))  # (bs * n, dx)
        true_E = torch.reshape(true_E, (-1, true_E.size(-1)))  # (bs * n * n, de)
        masked_pred_X = torch.reshape(pred.X, (-1, pred.X.size(-1)))  # (bs * n, dx)
        masked_pred_E = torch.reshape(pred.E, (-1, pred.E.size(-1)))   # (bs * n * n, de)

        # Remove masked rows
        mask_X = (true_X != 0.).any(dim=-1)
        mask_E = (true_E != 0.).any(dim=-1)

        flat_true_X = true_X[mask_X, :]
        flat_pred_X = masked_pred_X[mask_X, :]
        flat_true_E = true_E[mask_E, :]
        flat_pred_E = masked_pred_E[mask_E, :]
        
        # Calculate node and edge losses using cross entropy
        loss_X = F.cross_entropy(flat_pred_X, torch.argmax(flat_true_X, dim=-1)) if true_X.numel() > 0 else 0.0
        loss_E = F.cross_entropy(flat_pred_E, torch.argmax(flat_true_E, dim=-1)) if true_E.numel() > 0 else 0.0
        loss = lw_X * loss_X + lw_E * loss_E
        return loss, loss_X, loss_E

class AttentionBlock(nn.Module):
    def __init__(self, hidden_size, num_head, mlp_ratio=4.0, dropout=0.):
        super().__init__()
        self.dropout = dropout
        self.norm1 = nn.LayerNorm(hidden_size, elementwise_affine=False)
        self.norm2 = nn.LayerNorm(hidden_size, elementwise_affine=False)

        self.attn = AttentionWithNodeMask(
            hidden_size, num_head=num_head, qkv_bias=True, qk_norm=True, proj_drop=dropout
        )
        self.mlp = MLP(
            in_features=hidden_size,
            hidden_features=int(hidden_size * mlp_ratio),
            drop=self.dropout,
            use_bn=False,
        )
        self.adaLN_modulation = nn.Sequential(
            nn.Linear(hidden_size, hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(hidden_size, 6 * hidden_size, bias=True)
        )
        
    def forward(self, x, c, node_mask):
        (
            shift_msa,
            scale_msa,
            gate_msa,
            shift_mlp,
            scale_mlp,
            gate_mlp,
        ) = self.adaLN_modulation(c).chunk(6, dim=1)
        x = x + gate_msa.unsqueeze(1) * modulate(self.norm1(self.attn(x, node_mask=node_mask)), shift_msa, scale_msa)
        x = x + gate_mlp.unsqueeze(1) * modulate(self.norm2(self.mlp(x)), shift_mlp, scale_mlp)
        return x

class FinalLayer(nn.Module):
    def __init__(self, max_node, hidden_size, atom_type, bond_type, mlp_ratio, num_head=None):
        super().__init__()
        self.atom_type = atom_type
        self.bond_type = bond_type
        final_size = atom_type + max_node * bond_type
        self.XEdecoder = MLP(in_features=hidden_size, out_features=final_size, drop=0., use_bn=False)

        self.norm_final = nn.LayerNorm(final_size, elementwise_affine=False)
        self.adaLN_modulation = nn.Sequential(
            nn.Linear(hidden_size, hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(hidden_size, 2 * final_size, bias=True)
        )

    def forward(self, x, x_in, e_in, c, t, node_mask):
        x_all = self.XEdecoder(x)
        bs, n, _ = x_all.size()
        shift, scale = self.adaLN_modulation(c).chunk(2, dim=1)
        x_all = modulate(self.norm_final(x_all), shift, scale)
        
        atom_out = x_all[:, :, :self.atom_type]
        atom_out = x_in + atom_out

        bond_out = x_all[:, :, self.atom_type:].reshape(bs, n, n, self.bond_type)
        bond_out = e_in + bond_out

        ##### standardize adj_out
        edge_mask = (~node_mask)[:, :, None] & (~node_mask)[:, None, :]
        diag_mask = (
            torch.eye(n, dtype=torch.bool)
            .unsqueeze(0)
            .expand(bs, -1, -1)
            .type_as(edge_mask)
        )
        bond_out.masked_fill_(edge_mask[:, :, :, None], 0)
        bond_out.masked_fill_(diag_mask[:, :, :, None], 0)
        bond_out = 1 / 2 * (bond_out + torch.transpose(bond_out, 1, 2))
        return atom_out, bond_out, None


import torch
import torch.nn.functional as F

from .layers import DenseGCNConv, MLP, AttentionLayer
from .utils import mask_x, pow_tensor

class ScoreNetworkX(torch.nn.Module):
    def __init__(self, max_feat_num, depth, nhid):
        super(ScoreNetworkX, self).__init__()

        self.nfeat = max_feat_num
        self.depth = depth
        self.nhid = nhid

        self.layers = torch.nn.ModuleList()
        for _ in range(self.depth):
            if _ == 0:
                self.layers.append(DenseGCNConv(self.nfeat, self.nhid))
            else:
                self.layers.append(DenseGCNConv(self.nhid, self.nhid))

        self.fdim = self.nfeat + self.depth * self.nhid
        self.final = MLP(num_layers=3, input_dim=self.fdim, hidden_dim=2*self.fdim, output_dim=self.nfeat, 
                            use_bn=False, activate_func=F.elu)

        self.activation = torch.tanh

    def forward(self, x, adj, flags):

        x_list = [x]
        for _ in range(self.depth):
            x = self.layers[_](x, adj)
            x = self.activation(x)
            x_list.append(x)

        xs = torch.cat(x_list, dim=-1) # B x N x (F + num_layers x H)
        out_shape = (adj.shape[0], adj.shape[1], -1)
        x = self.final(xs).view(*out_shape)

        x = mask_x(x, flags)

        return x


# class ScoreNetworkX_GMH(torch.nn.Module):
#     def __init__(self, max_feat_num, depth, nhid, num_linears,
#                  c_init, c_hid, c_final, adim, num_heads=4, conv='GCN'):
#         super().__init__()

#         self.depth = depth
#         self.c_init = c_init

#         self.layers = torch.nn.ModuleList()
#         for _ in range(self.depth):
#             if _ == 0:
#                 self.layers.append(AttentionLayer(num_linears, max_feat_num, nhid, nhid, c_init, 
#                                                   c_hid, num_heads, conv))
#             elif _ == self.depth - 1:
#                 self.layers.append(AttentionLayer(num_linears, nhid, adim, nhid, c_hid, 
#                                                   c_final, num_heads, conv))
#             else:
#                 self.layers.append(AttentionLayer(num_linears, nhid, adim, nhid, c_hid, 
#                                                   c_hid, num_heads, conv))

#         fdim = max_feat_num + depth * nhid
#         self.final = MLP(num_layers=3, input_dim=fdim, hidden_dim=2*fdim, output_dim=max_feat_num, 
#                          use_bn=False, activate_func=F.elu)

#         self.activation = torch.tanh

#     def forward(self, x, adj, flags):
#         adjc = pow_tensor(adj, self.c_init)

#         x_list = [x]
#         for _ in range(self.depth):
#             x, adjc = self.layers[_](x, adjc, flags)
#             x = self.activation(x)
#             x_list.append(x)

#         xs = torch.cat(x_list, dim=-1) # B x N x (F + num_layers x H)
#         out_shape = (adj.shape[0], adj.shape[1], -1)
#         x = self.final(xs).view(*out_shape)
#         x = mask_x(x, flags)

#         return x


import os
import numpy as np
import warnings
import datetime
from tqdm import tqdm
from typing import Optional, Union, Dict, Any, Tuple, List, Callable, Literal, Type
from dataclasses import dataclass, field

import torch
from torch_geometric.loader import DataLoader

from .model import DIR
from ..gnn.modeling_gnn import GNNMolecularPredictor
from ...utils.search import (
    ParameterSpec,
    ParameterType,
)

@dataclass
class DIRMolecularPredictor(GNNMolecularPredictor):
    """This predictor implements the DIR for molecular property prediction tasks.

    The full name of DIR is Discovering Invariant Rationales.

    References
    ----------
    - Discovering Invariant Rationales for Graph Neural Networks.
      https://openreview.net/forum?id=hGXij5rfiHw
    - Code: https://github.com/Wuyxin/DIR-GNN
        
    Parameters
    ----------
    causal_ratio : float, default=0.8
        The ratio of causal edges to keep during training. A higher ratio means more edges
        are considered causal/important for the prediction. This controls the sparsity of
        the learned rationales.
        
    lw_invariant : float, default=1e-4
        The weight of the invariance loss term. This loss encourages the model to learn
        rationales that are invariant across different environments/perturbations. A higher
        value puts more emphasis on learning invariant features.
    """
    
    # DIR-specific parameters
    causal_ratio: float = 0.8
    lw_invariant: float = 1e-4
    # Override parent defaults
    model_name: str = "DIRMolecularPredictor"
    model_class: Type[DIR] = field(default=DIR, init=False)
    
    def __post_init__(self):
        super().__post_init__()

    @staticmethod
    def _get_param_names() -> List[str]:
        return ["causal_ratio", "lw_invariant"] + GNNMolecularPredictor._get_param_names()

    def _get_default_search_space(self):
        search_space = super()._get_default_search_space().copy()
        search_space["causal_ratio"] = ParameterSpec(ParameterType.FLOAT, (0.1, 0.9))
        search_space["lw_invariant"] = ParameterSpec(ParameterType.FLOAT, (1e-5, 1e-2))
        return search_space

    def _setup_optimizers(self) -> Tuple[Dict[str, torch.optim.Optimizer], Optional[Any]]:
        model_optimizer = torch.optim.Adam(self.model.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay)
        conf_optimizer = torch.optim.Adam(self.model.conf_lin.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay)
        if self.grad_clip_value is not None:
            for group in model_optimizer.param_groups:
                group.setdefault("max_norm", self.grad_clip_value)
                group.setdefault("norm_type", 2.0)

        scheduler = None
        if self.use_lr_scheduler:
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                model_optimizer,
                mode="min",
                factor=self.scheduler_factor,
                patience=self.scheduler_patience,
                min_lr=1e-6,
                cooldown=0,
                eps=1e-8,
            )
        optimizer = {"model": model_optimizer, "conf": conf_optimizer}

        return optimizer, scheduler
    
    def _get_model_params(self, checkpoint: Optional[Dict] = None) -> Dict[str, Any]:
        base_params = super()._get_model_params(checkpoint)
        if checkpoint and "hyperparameters" in checkpoint:
            base_params["causal_ratio"] = checkpoint["hyperparameters"]["causal_ratio"]
        else:
            base_params["causal_ratio"] = self.causal_ratio
        return base_params

    def _train_epoch(self, train_loader, optimizer, epoch):
        self.model.train()
        losses = []

        iterator = (
            tqdm(train_loader, desc="Training", leave=False)
            if self.verbose
            else train_loader
        )

        alpha_prime = self.lw_invariant * (epoch ** 1.6)
        conf_opt = optimizer["conf"]
        model_optimizer = optimizer["model"]

        for batch in iterator:

            batch = batch.to(self.device)

            # Forward pass and loss computation
            causal_loss, conf_loss, env_loss = self.model.compute_loss(batch, self.loss_criterion, alpha_prime)

            conf_opt.zero_grad()
            conf_loss.backward()
            conf_opt.step()

            model_optimizer.zero_grad()
            (causal_loss + env_loss).backward()
            model_optimizer.step()

            loss = causal_loss + env_loss + conf_loss
            losses.append(loss.item())

            # Update progress bar if using tqdm
            if self.verbose:
                iterator.set_postfix({"Epoch": epoch, "Causal Loss": f"{causal_loss.item():.4f}", "Conf Loss": f"{conf_loss.item():.4f}", "Env Loss": f"{env_loss.item():.4f}", "Total Loss": f"{loss.item():.4f}"})

        return losses
    
    def predict(self, X: List[str]) -> Dict[str, Union[np.ndarray, List[List]]]:
        self._check_is_fitted()

        # Convert to PyTorch Geometric format and create loader
        X, _ = self._validate_inputs(X)
        dataset = self._convert_to_pytorch_data(X)
        loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=False)

        # Make predictions
        self.model = self.model.to(self.device)
        self.model.eval()
        predictions = []
        with torch.no_grad():
            for batch in tqdm(loader, disable=not self.verbose):
                batch = batch.to(self.device)
                out = self.model(batch)
                predictions.append(out["prediction"].cpu().numpy())

        if predictions:
            return {
                "prediction": np.concatenate(predictions, axis=0),
            }
        else:
            warnings.warn(
                "No valid predictions could be made from the input data. Returning empty results."
            )
            return {"prediction": np.array([])}





import torch
from torch.utils.data import Dataset, DataLoader
from .mol_tree import MolTree
import numpy as np
from .jtnn_enc import JTNNEncoder
from .mpn import MPN
from .jtmpn import JTMPN
import pickle
import os, random

# class PairTreeFolder(object):
#     def __init__(self, data_folder, vocab, batch_size, num_workers=4, shuffle=True, y_assm=True, replicate=None):
#         self.data_folder = data_folder
#         self.data_files = [fn for fn in os.listdir(data_folder)]
#         self.batch_size = batch_size
#         self.vocab = vocab
#         self.num_workers = num_workers
#         self.y_assm = y_assm
#         self.shuffle = shuffle

#         if replicate is not None: #expand is int
#             self.data_files = self.data_files * replicate

#     def __iter__(self):
#         for fn in self.data_files:
#             fn = os.path.join(self.data_folder, fn)
#             with open(fn) as f:
#                 data = pickle.load(f)

#             if self.shuffle: 
#                 random.shuffle(data) #shuffle data before batch

#             batches = [data[i : i + self.batch_size] for i in range(0, len(data), self.batch_size)]
#             if len(batches[-1]) < self.batch_size:
#                 batches.pop()

#             dataset = PairTreeDataset(batches, self.vocab, self.y_assm)
#             dataloader = DataLoader(dataset, batch_size=1, shuffle=False, num_workers=self.num_workers, collate_fn=lambda x:x[0])

#             for b in dataloader:
#                 yield b

#             del data, batches, dataset, dataloader

class MolTreeFolder(object):
    def __init__(self, data, vocab, batch_size, num_workers=0, shuffle=True, assm=True):
        self.data = data
        self.batch_size = batch_size
        self.vocab = vocab
        self.num_workers = num_workers
        self.shuffle = shuffle
        self.assm = assm

    def __iter__(self):
        data = self.data
        if self.shuffle: 
            random.shuffle(data) # shuffle data before batch

        batches = [data[i : i + self.batch_size] for i in range(0, len(data), self.batch_size)]
        if len(batches[-1]) < self.batch_size:
            batches.pop()

        dataset = MolTreeDataset(batches, self.vocab, self.assm)
        dataloader = DataLoader(dataset, batch_size=1, shuffle=False, num_workers=self.num_workers, collate_fn=lambda x:x[0])

        for b in dataloader:
            yield b

        del data, batches, dataset, dataloader

class PairTreeDataset(Dataset):

    def __init__(self, data, vocab, y_assm):
        self.data = data
        self.vocab = vocab
        self.y_assm = y_assm

    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        batch0, batch1 = zip(*self.data[idx])
        return tensorize(batch0, self.vocab, assm=False), tensorize(batch1, self.vocab, assm=self.y_assm)

class MolTreeDataset(Dataset):
    def __init__(self, data, vocab, assm=True):
        self.data = data
        self.vocab = vocab
        self.assm = assm

    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        return tensorize(self.data[idx], self.vocab, assm=self.assm)

def tensorize(tree_batch, vocab, assm=True):
    set_batch_nodeID(tree_batch, vocab)
    smiles_batch = [tree.smiles for tree in tree_batch]
    jtenc_holder,mess_dict = JTNNEncoder.tensorize(tree_batch)
    jtenc_holder = jtenc_holder
    mpn_holder = MPN.tensorize(smiles_batch)

    if assm is False:
        return tree_batch, jtenc_holder, mpn_holder

    cands = []
    batch_idx = []
    for i,mol_tree in enumerate(tree_batch):
        for node in mol_tree.nodes:
            #Leaf node's attachment is determined by neighboring node's attachment
            if node.is_leaf or len(node.cands) == 1: continue
            cands.extend( [(cand, mol_tree.nodes, node) for cand in node.cands] )
            batch_idx.extend([i] * len(node.cands))

    jtmpn_holder = JTMPN.tensorize(cands, mess_dict)
    batch_idx = torch.LongTensor(batch_idx)

    return tree_batch, jtenc_holder, mpn_holder, (jtmpn_holder,batch_idx)

def set_batch_nodeID(mol_batch, vocab):
    tot = 0
    for mol_tree in mol_batch:
        for node in mol_tree.nodes:
            node.idx = tot
            node.wid = vocab.get_index(node.smiles)
            tot += 1



import random

import numpy as np
from rdkit import Chem, rdBase
from rdkit.Chem import AllChem
rdBase.DisableLog('rdApp.error')


def cut(mol):
    if not mol.HasSubstructMatch(Chem.MolFromSmarts('[*]-;!@[*]')):
        return None

    bis = random.choice(mol.GetSubstructMatches(Chem.MolFromSmarts('[*]-;!@[*]')))  # single bond not in ring

    bs = [mol.GetBondBetweenAtoms(bis[0], bis[1]).GetIdx()]

    fragments_mol = Chem.FragmentOnBonds(mol, bs, addDummies=True, dummyLabels=[(1, 1)])

    try:
        return Chem.GetMolFrags(fragments_mol, asMols=True, sanitizeFrags=True)
    except ValueError:
        return None

    return None


def cut_ring(mol):

    for i in range(10):
        if random.random() < 0.5:
            if not mol.HasSubstructMatch(Chem.MolFromSmarts('[R]@[R]@[R]@[R]')):
                return None
            bis = random.choice(mol.GetSubstructMatches(Chem.MolFromSmarts('[R]@[R]@[R]@[R]')))
            bis = ((bis[0], bis[1]), (bis[2], bis[3]),)
        else:
            if not mol.HasSubstructMatch(Chem.MolFromSmarts('[R]@[R;!D2]@[R]')):
                return None
            bis = random.choice(mol.GetSubstructMatches(Chem.MolFromSmarts('[R]@[R;!D2]@[R]')))
            bis = ((bis[0], bis[1]), (bis[1], bis[2]),)

        bs = [mol.GetBondBetweenAtoms(x, y).GetIdx() for x, y in bis]

        fragments_mol = Chem.FragmentOnBonds(mol, bs, addDummies=True, dummyLabels=[(1, 1), (1, 1)])

        try:
            fragments = Chem.GetMolFrags(fragments_mol, asMols=True, sanitizeFrags=True)
            if len(fragments) == 2:
                return fragments
        except ValueError:
            return None

    return None


def ring_OK(mol):
    if not mol.HasSubstructMatch(Chem.MolFromSmarts('[R]')):
        return True

    ring_allene = mol.HasSubstructMatch(Chem.MolFromSmarts('[R]=[R]=[R]'))

    cycle_list = mol.GetRingInfo().AtomRings()
    max_cycle_length = max([len(j) for j in cycle_list])
    macro_cycle = max_cycle_length > 6

    double_bond_in_small_ring = mol.HasSubstructMatch(Chem.MolFromSmarts('[r3,r4]=[r3,r4]'))

    return not ring_allene and not macro_cycle and not double_bond_in_small_ring


# TODO: set from main? calculate for dataset?
average_size = 39.15
size_stdev = 3.50


def mol_ok(mol):
    try:
        Chem.SanitizeMol(mol)
        target_size = size_stdev * np.random.randn() + average_size  # parameters set in GA_mol
        if mol.GetNumAtoms() > 5 and mol.GetNumAtoms() < target_size:
            return True
        else:
            return False
    except ValueError:
        return False


def crossover_ring(parent_A, parent_B):
    ring_smarts = Chem.MolFromSmarts('[R]')
    if not parent_A.HasSubstructMatch(ring_smarts) and not parent_B.HasSubstructMatch(ring_smarts):
        return None

    rxn_smarts1 = ['[*:1]~[1*].[1*]~[*:2]>>[*:1]-[*:2]', '[*:1]~[1*].[1*]~[*:2]>>[*:1]=[*:2]']
    rxn_smarts2 = ['([*:1]~[1*].[1*]~[*:2])>>[*:1]-[*:2]', '([*:1]~[1*].[1*]~[*:2])>>[*:1]=[*:2]']

    for i in range(10):
        fragments_A = cut_ring(parent_A)
        fragments_B = cut_ring(parent_B)

        if fragments_A is None or fragments_B is None:
            return None

        new_mol_trial = []
        for rs in rxn_smarts1:
            rxn1 = AllChem.ReactionFromSmarts(rs)
            new_mol_trial = []
            for fa in fragments_A:
                for fb in fragments_B:
                    if len(rxn1.RunReactants((fa, fb))) > 0:
                        new_mol_trial.append(rxn1.RunReactants((fa, fb))[0])

        new_mols = []
        for rs in rxn_smarts2:
            rxn2 = AllChem.ReactionFromSmarts(rs)
            for m in new_mol_trial:
                m = m[0]
                if mol_ok(m):
                    new_mols += list(rxn2.RunReactants((m,)))

        new_mols2 = []
        for m in new_mols:
            m = m[0]
            if mol_ok(m) and ring_OK(m):
                new_mols2.append(m)

        if len(new_mols2) > 0:
            return random.choice(new_mols2)

    return None


def crossover_non_ring(parent_A, parent_B):

    for i in range(10):
        fragments_A = cut(parent_A)
        fragments_B = cut(parent_B)
        if fragments_A is None or fragments_B is None:
            return None
        rxn = AllChem.ReactionFromSmarts('[*:1]-[1*].[1*]-[*:2]>>[*:1]-[*:2]')
        new_mol_trial = []
        for fa in fragments_A:
            for fb in fragments_B:
                if len(rxn.RunReactants((fa, fb))) > 0:
                    new_mol_trial.append(rxn.RunReactants((fa, fb))[0])

        new_mols = []
        for mol in new_mol_trial:
            mol = mol[0]
            if mol_ok(mol):
                new_mols.append(mol)

        if len(new_mols) > 0:
            return random.choice(new_mols)

    return None


def crossover(parent_A, parent_B):
    parent_smiles = [Chem.MolToSmiles(parent_A), Chem.MolToSmiles(parent_B)]
    try:
        Chem.Kekulize(parent_A, clearAromaticFlags=True)
        Chem.Kekulize(parent_B, clearAromaticFlags=True)

    except ValueError:
        pass

    for i in range(10):
        if random.random() <= 0.5:
            # print 'non-ring crossover'
            new_mol = crossover_non_ring(parent_A, parent_B)
            if new_mol is not None:
                new_smiles = Chem.MolToSmiles(new_mol)
                if new_smiles is not None and new_smiles not in parent_smiles:
                    return new_mol
        else:
            # print 'ring crossover'
            new_mol = crossover_ring(parent_A, parent_B)
            if new_mol is not None:
                new_smiles = Chem.MolToSmiles(new_mol)
                if new_smiles is not None and new_smiles not in parent_smiles:
                    return new_mol

    return None


import torch
import torch
import torch.nn as nn
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool
from ...nn import GNN_node, GNN_node_Virtualnode, MLP
from ...utils import init_weights


class RPGNN(nn.Module):
    def __init__(
        self,
        num_task,
        num_layer,
        num_node_feature,
        fixed_size=32,
        num_perm=3,
        hidden_size=300,
        gnn_type="gin-virtual",
        drop_ratio=0.5,
        norm_layer="batch_norm",
        graph_pooling="mean",
        augmented_feature=['maccs', 'morgan']
    ):
        """
        Random Permutation Graph Neural Network (RPGNN)
        
        Args:
            num_task: Number of prediction tasks
            num_layer: Number of GNN layers
            num_node_feature: Number of node features
            fixed_size: Size of position encoding vectors
            num_perm: Number of random permutations to average over
            hidden_size: Hidden dimension size
            gnn_type: Type of GNN to use (gin, gin-virtual, gcn, gcn-virtual)
            drop_ratio: Dropout ratio
            norm_layer: Type of normalization layer
            graph_pooling: Type of graph pooling (mean, sum, max)
            augmented_feature: List of additional molecular features to use
        """
        super(RPGNN, self).__init__()
        
        gnn_name = gnn_type.split("-")[0]
        self.num_task = num_task
        self.hidden_size = hidden_size
        self.fixed_size = fixed_size
        self.num_perm = num_perm
        self.augmented_feature = augmented_feature
        self.num_node_feature = num_node_feature        
        # Register fixed-size node IDs (position encoding)
        self.register_buffer("node_ids", torch.eye(self.fixed_size))
        
        # GNN selection
        if "virtual" in gnn_type:
            self.gnn = GNN_node_Virtualnode(
                num_layer=num_layer,
                hidden_size=hidden_size,
                JK="last",
                drop_ratio=drop_ratio,
                residual=True,
                gnn_name=gnn_name,
                norm_layer=norm_layer
            )
        else:
            self.gnn = GNN_node(
                num_layer=num_layer,
                hidden_size=hidden_size,
                JK="last",
                drop_ratio=drop_ratio,
                residual=True,
                gnn_name=gnn_name,
                norm_layer=norm_layer
            )
        
        # Graph pooling selection
        if graph_pooling == "sum":
            self.pool = global_add_pool
        elif graph_pooling == "mean":
            self.pool = global_mean_pool
        elif graph_pooling == "max":
            self.pool = global_max_pool
        else:
            raise ValueError(f"Invalid graph pooling type: {graph_pooling}")
            
        # Calculate final prediction dimension
        graph_dim = hidden_size+self.fixed_size
        if augmented_feature:
            if "morgan" in augmented_feature:
                graph_dim += 1024
            if "maccs" in augmented_feature:
                graph_dim += 167
                
        # Final predictor MLP
        self.predictor = MLP(
            in_features=graph_dim,
            hidden_features=2 * hidden_size,
            out_features=num_task
        )

    def initialize_parameters(self, seed=None):
        """Initialize all model parameters."""
        if seed is not None:
            torch.manual_seed(seed)
            
        # Initialize GNN and predictor
        init_weights(self.gnn)
        init_weights(self.predictor)
        
        # Apply reset_parameters to all modules that have it
        def reset_parameters(module):
            if hasattr(module, 'reset_parameters'):
                module.reset_parameters()
        
        self.apply(reset_parameters)

    def _augmented_graph_features(self, batched_data, h_rep):
        """Add augmented molecular features to the graph representation."""
        if not self.augmented_feature:
            return h_rep
            
        if "morgan" in self.augmented_feature:
            morgan = batched_data.morgan.type_as(h_rep)
            h_rep = torch.cat((h_rep, morgan), dim=1)
        if "maccs" in self.augmented_feature:
            maccs = batched_data.maccs.type_as(h_rep)
            h_rep = torch.cat((h_rep, maccs), dim=1)
            
        return h_rep

    def forward(self, batched_data):
        """Forward pass of RPGNN."""
        # Extract original data
        x, edge_index, batch = batched_data.x, batched_data.edge_index, batched_data.batch
        
        # Initialize final output
        out = None
        
        # Average over multiple permutations
        for _ in range(self.num_perm):
            temp_data = batched_data.clone()
            # Create new_x with same dtype as x
            new_x = torch.empty(x.size(0), x.size(1) + self.fixed_size, dtype=x.dtype, device=x.device)
            
            for graph_idx in range(batch.max().item() + 1):
                node_indices = (batch == graph_idx).nonzero().squeeze(1)
                graph_size = node_indices.size(0)
                perm = torch.randperm(graph_size)

                node_ids = self.node_ids.repeat(
                    graph_size // self.fixed_size + 1, 1
                )[:graph_size]
                permuted_node_ids = node_ids[perm, :]
                
                # Convert both tensors to the same type as x before concatenation
                new_x[node_indices] = torch.cat([x[node_indices].to(x.dtype), permuted_node_ids.to(x.dtype)], dim=1)
            temp_data.x = new_x[:,:self.num_node_feature]
            # Process through GNN
            h_v, _ = self.gnn(temp_data)

            fused_h_v = torch.cat([h_v, new_x[:,self.num_node_feature:]], dim=1)

            # Pool node representations to graph representations
            h_rep = self.pool(fused_h_v, temp_data.batch)
            
            # Add augmented features
            h_rep = self._augmented_graph_features(temp_data, h_rep)
            
            # Accumulate output
            if out is None:
                out = h_rep / self.num_perm
            else:
                out += h_rep / self.num_perm
        
        # Final prediction
        prediction = self.predictor(out)
        
        return {"prediction": prediction}

    def compute_loss(self, batched_data, criterion):
        out = self.forward(batched_data)
        prediction = out['prediction']
        target = batched_data.y.to(torch.float32)
        is_labeled = batched_data.y == batched_data.y
        loss = criterion(prediction.to(torch.float32)[is_labeled], target[is_labeled]).mean()
        
        return loss


import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import deque
from .mol_tree import Vocab, MolTree
from .nnutils import create_var, index_select_ND

class JTNNEncoder(nn.Module):

    def __init__(self, hidden_size, depth, embedding):
        super(JTNNEncoder, self).__init__()
        self.hidden_size = hidden_size
        self.depth = depth

        self.embedding = embedding
        self.outputNN = nn.Sequential(
            nn.Linear(2 * hidden_size, hidden_size),
            nn.ReLU()
        )
        self.GRU = GraphGRU(hidden_size, hidden_size, depth=depth)

    def forward(self, fnode, fmess, node_graph, mess_graph, scope):
        fnode = create_var(fnode)
        fmess = create_var(fmess)
        node_graph = create_var(node_graph)
        mess_graph = create_var(mess_graph)
        messages = create_var(torch.zeros(mess_graph.size(0), self.hidden_size))

        fnode = self.embedding(fnode)
        fmess = index_select_ND(fnode, 0, fmess)
        messages = self.GRU(messages, fmess, mess_graph)

        mess_nei = index_select_ND(messages, 0, node_graph)
        node_vecs = torch.cat([fnode, mess_nei.sum(dim=1)], dim=-1)
        node_vecs = self.outputNN(node_vecs)

        max_len = max([x for _,x in scope])
        batch_vecs = []
        for st,le in scope:
            cur_vecs = node_vecs[st] #Root is the first node
            batch_vecs.append( cur_vecs )

        tree_vecs = torch.stack(batch_vecs, dim=0)
        return tree_vecs, messages

    @staticmethod
    def tensorize(tree_batch):
        node_batch = [] 
        scope = []
        for tree in tree_batch:
            scope.append( (len(node_batch), len(tree.nodes)) )
            node_batch.extend(tree.nodes)

        return JTNNEncoder.tensorize_nodes(node_batch, scope)
    
    @staticmethod
    def tensorize_nodes(node_batch, scope):
        messages,mess_dict = [None],{}
        fnode = []
        for x in node_batch:
            fnode.append(x.wid)
            for y in x.neighbors:
                mess_dict[(x.idx,y.idx)] = len(messages)
                messages.append( (x,y) )

        node_graph = [[] for i in range(len(node_batch))]
        mess_graph = [[] for i in range(len(messages))]
        fmess = [0] * len(messages)

        for x,y in messages[1:]:
            mid1 = mess_dict[(x.idx,y.idx)]
            fmess[mid1] = x.idx 
            node_graph[y.idx].append(mid1)
            for z in y.neighbors:
                if z.idx == x.idx: continue
                mid2 = mess_dict[(y.idx,z.idx)]
                mess_graph[mid2].append(mid1)

        max_len = max([len(t) for t in node_graph] + [1])
        for t in node_graph:
            pad_len = max_len - len(t)
            t.extend([0] * pad_len)

        max_len = max([len(t) for t in mess_graph] + [1])
        for t in mess_graph:
            pad_len = max_len - len(t)
            t.extend([0] * pad_len)

        mess_graph = torch.LongTensor(mess_graph)
        node_graph = torch.LongTensor(node_graph)
        fmess = torch.LongTensor(fmess)
        fnode = torch.LongTensor(fnode)
        return (fnode, fmess, node_graph, mess_graph, scope), mess_dict

class GraphGRU(nn.Module):

    def __init__(self, input_size, hidden_size, depth):
        super(GraphGRU, self).__init__()
        self.hidden_size = hidden_size
        self.input_size = input_size
        self.depth = depth

        self.W_z = nn.Linear(input_size + hidden_size, hidden_size)
        self.W_r = nn.Linear(input_size, hidden_size, bias=False)
        self.U_r = nn.Linear(hidden_size, hidden_size)
        self.W_h = nn.Linear(input_size + hidden_size, hidden_size)

    def forward(self, h, x, mess_graph):
        mask = torch.ones(h.size(0), 1)
        mask[0] = 0 #first vector is padding
        mask = create_var(mask)
        for it in range(self.depth):
            h_nei = index_select_ND(h, 0, mess_graph)
            sum_h = h_nei.sum(dim=1)
            z_input = torch.cat([x, sum_h], dim=1)
            z = F.sigmoid(self.W_z(z_input))

            r_1 = self.W_r(x).view(-1, 1, self.hidden_size)
            r_2 = self.U_r(h_nei)
            r = F.sigmoid(r_1 + r_2)
            
            gated_h = r * h_nei
            sum_gated_h = gated_h.sum(dim=1)
            h_input = torch.cat([x, sum_gated_h], dim=1)
            pre_h = F.tanh(self.W_h(h_input))
            h = (1.0 - z) * sum_h + z * pre_h
            h = h * mask

        return h





import os
import numpy as np
from tqdm import tqdm
from typing import Optional, Union, Dict, Any, Tuple, List
import warnings
import datetime

import torch
from torch_geometric.loader import DataLoader
from torch_geometric.data import Data

from .graph_utils.features import smiles_to_graph
from ..base import BaseMolecularPredictor

class MolecularGraphPredictor(BaseMolecularPredictor):
    """This predictor implements a GNN model for molecular property prediction tasks.
    """

    def __init__(
        self,
        num_tasks: int = 1,
        task_type: str = "classification",
        num_layer: int = 5,
        emb_dim: int = 300,
        gnn_type: str = "gin-virtual",
        drop_ratio: float = 0.5,
        norm_layer: str = "batch_norm",
        batch_size: int = 32,
        epochs: int = 100,
        learning_rate: float = 0.001,
        weight_decay: float = 0.0,
        patience: int = 50,
        device: Optional[str] = None,
        use_lr_scheduler: bool = True,
        scheduler_factor: float = 0.5,
        scheduler_patience: int = 5,
        grad_clip_value: Optional[float] = None,
        verbose: bool = False,
        model_name: str = "graph_predictor",
        criterion = None,
    ):
        super().__init__(num_tasks=num_tasks, task_type=task_type)

        # Model hyperparameters
        self.num_layer = num_layer
        self.emb_dim = emb_dim
        self.gnn_type = gnn_type
        self.drop_ratio = drop_ratio
        self.norm_layer = norm_layer

        # Training parameters
        self.batch_size = batch_size
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.patience = patience
        self.use_lr_scheduler = use_lr_scheduler
        self.scheduler_factor = scheduler_factor
        self.scheduler_patience = scheduler_patience
        self.grad_clip_value = grad_clip_value
        self.criterion = criterion if criterion is not None else self._load_default_criterion()

        # Set device
        self.device = (
            torch.device(device)
            if device
            else torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        )

        # Initialize model
        self.model = None

        self.fitting_loss = []
        self.fitting_loss_mean = float("inf")
        self.fitting_epoch = 0
        self.verbose = verbose
        self.model_name = model_name

    @staticmethod
    def _get_param_names() -> List[str]:
        """Get parameter names for the estimator.

        Returns
        -------
        List[str]
            List of parameter names that can be used for model configuration.
        """
        return [
            # Model hyperparameters
            "num_tasks",
            "task_type",
            "num_layer",
            "emb_dim",
            "gnn_type",
            "drop_ratio",
            "norm_layer",
            # Training parameters
            "batch_size",
            "epochs",
            "learning_rate",
            "weight_decay",
            "patience",
            "device",
            "grad_clip_value",
            "criterion",
            # Scheduler parameters
            "use_lr_scheduler",
            "scheduler_factor",
            "scheduler_patience",
        ]

    def _get_search_parameters(self) -> Dict[str, Union[Tuple[float, float], Tuple[int, int]]]:
        """Get the default hyperparameter search space for auto_fit.

        Returns
        -------
        Dict[str, Union[Tuple[float, float], Tuple[int, int]]]
            Dictionary containing parameter names and their search ranges:
            - num_layer: Number of GNN layers (2 to 8)
            - emb_dim: Embedding dimension (64 to 512)
            - drop_ratio: Dropout ratio (0.0 to 0.7)
            - learning_rate: Learning rate for optimization (1e-4 to 1e-2)
            - weight_decay: L2 regularization factor (1e-6 to 1e-3)
        """
        return {
            'num_layer': (2, 8),
            'emb_dim': (64, 512),
            'drop_ratio': (0.0, 0.7),
            'learning_rate': (1e-4, 1e-2),
            'weight_decay': (1e-6, 1e-3),
        }

    def _convert_to_pytorch_data(self, X, y=None):
        """Convert numpy arrays to PyTorch data format.

        Should be implemented by child classes to handle specific molecular
        representations (e.g., SMILES to graphs, conformers to graphs, etc.)
        """
        if self.verbose:
            iterator = tqdm(enumerate(X), desc="Converting molecules to graphs", total=len(X))
        else:
            iterator = enumerate(X)

        pyg_graph_list = []
        for idx, smiles in iterator:
            if y is not None:
                properties = y[idx]
            else:
                properties = None
            graph = smiles_to_graph(smiles, properties)
            g = Data()
            g.num_nodes = graph["num_nodes"]
            g.edge_index = torch.from_numpy(graph["edge_index"])

            del graph["num_nodes"]
            del graph["edge_index"]

            if graph["edge_feat"] is not None:
                g.edge_attr = torch.from_numpy(graph["edge_feat"])
                del graph["edge_feat"]

            if graph["node_feat"] is not None:
                g.x = torch.from_numpy(graph["node_feat"])
                del graph["node_feat"]

            if graph["y"] is not None:
                g.y = torch.from_numpy(graph["y"])
                del graph["y"]

            try:
                g.fp = torch.tensor(graph["fp"], dtype=torch.int8).view(1, -1)
                del graph["fp"]
            except:
                pass

            pyg_graph_list.append(g)

        return pyg_graph_list
    
    def setup_optimizers(self) -> Tuple[torch.optim.Optimizer, Optional[Any]]:
        """Setup optimization components including optimizer and learning rate scheduler.

        Returns
        -------
        Tuple[optim.Optimizer, Optional[Any]]
            A tuple containing:
            - The configured optimizer
            - The learning rate scheduler (if enabled, else None)

        Notes
        -----
        The scheduler returned can be either _LRScheduler or ReduceLROnPlateau,
        hence we use Any as the type hint instead of optim.lr_scheduler._LRScheduler
        """
        optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay
        )
        if self.grad_clip_value is not None:
            for group in optimizer.param_groups:
                group.setdefault("max_norm", self.grad_clip_value)
                group.setdefault("norm_type", 2.0)

        scheduler = None
        if self.use_lr_scheduler:
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                optimizer,
                mode="min",
                factor=self.scheduler_factor,
                patience=self.scheduler_patience,
                min_lr=1e-6,
                cooldown=0,
                eps=1e-8,
            )

        return optimizer, scheduler
    
    def auto_fit(
        self,
        X_train,
        y_train,
        search_parameters: dict = None,
        n_trials: int = 10,
        metric: str = "loss",
        X_val=None,
        y_val=None,
    ) -> "MolecularGraphPredictor":
        """Automatically find the best hyperparameters using Optuna optimization.
        
        Parameters
        ----------
        X_train : List[str]
            Training data as SMILES strings
        y_train : Union[List, np.ndarray]
            Training labels
        X_val : List[str], optional
            Validation data
        y_val : Union[List, np.ndarray], optional
            Validation labels
        search_parameters : dict, optional
            Dictionary of parameters to search over
        n_trials : int, default=100
            Number of optimization trials
        metric : str, default="rmse"
            Metric to optimize ("rmse", "mae", "r2", or "loss")
            
        Returns
        -------
        self : MolecularGraphPredictor
            Fitted estimator
        """
        import optuna
        
        # Default search parameters
        if search_parameters is None:
            search_parameters = self._get_search_parameters()
        else:
            # Validate search parameters against allowed parameters
            valid_params = set(self._get_param_names())
            invalid_params = set(search_parameters.keys()) - valid_params
            if invalid_params:
                raise ValueError(
                    f"Invalid search parameters: {invalid_params}. "
                    f"Valid parameters are: {valid_params}"
                )
        
        # Variables to track best state
        best_score = float('inf')
        best_param_vector = None
        best_trial_params = None
        best_loss = None
        best_epoch = None
        
        def objective(trial):
            nonlocal best_score, best_param_vector, best_trial_params, best_loss, best_epoch
            
            # Define hyperparameters to optimize
            params = {}
            for param, value_range in search_parameters.items():
                if isinstance(value_range, tuple):
                    if param in ['learning_rate', 'weight_decay']:
                        params[param] = trial.suggest_float(
                            param, value_range[0], value_range[1], log=True
                        )
                    elif param in ['drop_ratio', 'gamma']:
                        params[param] = trial.suggest_float(
                            param, value_range[0], value_range[1]
                        )
                    else:
                        params[param] = trial.suggest_int(
                            param, value_range[0], value_range[1]
                        )
                elif isinstance(value_range, list):
                    params[param] = trial.suggest_categorical(param, value_range)
            
            # Update model parameters
            self.set_params(**params)
            
            try:
                # Train model
                self.fit(X_train, y_train, X_val, y_val)

                if metric == "loss":
                    score = self.fitting_loss_mean
                else:
                    y_eval = y_train
                    X_eval = X_train
                    eval_results = self.predict(X_eval)
                    y_pred = eval_results['prediction']
                    
                    # Calculate metric
                    if metric == "rmse":
                        from sklearn.metrics import mean_squared_error
                        score = mean_squared_error(y_eval, y_pred, squared=False)
                    elif metric == "mae":
                        from sklearn.metrics import mean_absolute_error
                        score = mean_absolute_error(y_eval, y_pred)
                    elif metric == "r2":
                        from sklearn.metrics import r2_score
                        score = -r2_score(y_eval, y_pred)  # Negative because Optuna minimizes
                    else:
                        raise ValueError(f"Unknown metric: {metric}")
                
                # Update best state if current score is better
                if score < best_score:
                    best_score = score
                    best_param_vector = torch.nn.utils.parameters_to_vector(self.model.parameters())
                    best_trial_params = params
                    best_loss = self.fitting_loss
                    best_epoch = self.fitting_epoch
                
                print(f'done {score}', self)
                return score
                
            except Exception as e:
                print(f"Trial failed with error: {str(e)}")
                return float('inf')
        
        # Create study
        study = optuna.create_study(direction="minimize")
        
        # Run optimization
        study.optimize(
            objective,
            n_trials=n_trials,
            catch=(Exception,),
            show_progress_bar=True if self.verbose else False
        )
        
        # Load best parameters and state
        if best_param_vector is not None:
            # Update model hyperparameters
            self.set_params(**best_trial_params)
            # Load best model parameters
            torch.nn.utils.vector_to_parameters(best_param_vector, self.model.parameters())
            # Update training state
            self.fitting_loss = best_loss
            self.fitting_epoch = best_epoch
            self.fitting_loss_mean = float(np.mean(best_loss))
            self.is_fitted_ = True
        else:
            raise RuntimeError("No successful trials completed during optimization")
        
        return self
    
    def fit(self, X_train: List[str], y_train: Optional[Union[List, np.ndarray]], X_val=None, y_val=None) -> "MolecularGraphPredictor":
        """Fit the model to the data.

        Parameters
        ----------
        X_train / X_val : list of SMILES strings
            Input molecular structures as SMILES strings
        y_train / y_val : array-like,
            Target values for property prediction
        Returns
        -------
        self : object
            Fitted estimator
        """
        if self.model is None:
            raise RuntimeError("Model not initialized")

        self.model.initialize_parameters()
        self.model = self.model.to(self.device)
        # Setup training
        optimizer, scheduler = self.setup_optimizers()
        cnt_wait = 0

        # Convert molecular data to PyTorch Geometric format
        X_train, y_train = self._validate_inputs(X_train, y_train)
        dataset = self._convert_to_pytorch_data(X_train, y_train)
        loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=False)

        # Initialize best model state
        self.fitting_loss = []
        self.fitting_loss_mean = float("inf")
        self.fitting_epoch = 0
        best_param_vector = None

        # Training loop
        for epoch in range(self.epochs):
            # Train for one epoch
            losses = self.train_epoch(loader, optimizer)
            loss_mean = np.mean(losses)
            if scheduler:
                scheduler.step(loss_mean)

            # Update best model
            if loss_mean < self.fitting_loss_mean:
                self.fitting_epoch = epoch
                self.fitting_loss = losses
                self.fitting_loss_mean = loss_mean
                best_param_vector = torch.nn.utils.parameters_to_vector(self.model.parameters())
                cnt_wait = 0
            else:
                cnt_wait += 1
                if cnt_wait > self.patience:
                    break

        if best_param_vector is not None:
            torch.nn.utils.vector_to_parameters(best_param_vector, self.model.parameters())

        self.is_fitted_ = True
        return self

    def predict(self, X: List[str]) -> Dict[str, np.ndarray]:
        """Make predictions using the fitted model.

        Parameters
        ----------
        X : List[str]
            List of SMILES strings to make predictions for

        Returns
        -------
        Dict[str, np.ndarray]
            Dictionary containing:
                - 'prediction': Model predictions (shape: [n_samples, n_tasks])

        """
        self._check_is_fitted()

        # Convert to PyTorch Geometric format and create loader
        X, _ = self._validate_inputs(X)
        dataset = self._convert_to_pytorch_data(X)
        loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=False)

        if self.model is None:
            raise RuntimeError("Model not initialized")
        # Make predictions
        self.model = self.model.to(self.device)
        self.model.eval()
        predictions = []
        with torch.no_grad():
            for batch in loader:
                batch = batch.to(self.device)
                out = self.model(batch)
                predictions.append(out["prediction"].cpu().numpy())

        return {
            "prediction": np.concatenate(predictions, axis=0),
        }

    def train_epoch(self, train_loader, optimizer):
        """Training logic for one epoch.

        Args:
            train_loader: DataLoader containing training data
            optimizer: Optimizer instance for model parameter updates

        Returns:
            list: List of loss values for each training step
        """
        self.model.train()
        losses = []

        iterator = (
            tqdm(train_loader, desc="Training", leave=False)
            if self.verbose
            else train_loader
        )

        for batch in iterator:
            batch = batch.to(self.device)
            optimizer.zero_grad()

            # Forward pass and loss computation
            loss = self.model.compute_loss(batch, self.criterion)

            # Backward pass
            loss.backward()

            # Compute gradient norm if gradient clipping is enabled
            if self.grad_clip_value is not None:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip_value)

            optimizer.step()

            losses.append(loss.item())

            # Update progress bar if using tqdm
            if self.verbose:
                iterator.set_postfix({"loss": f"{loss.item():.4f}"})

        return losses

    def save_model(self, path: str) -> None:
        """Save the model to disk.

        Parameters
        ----------
        path : str
            Path where to save the model
        """
        if not self.is_fitted_:
            raise ValueError("Model must be fitted before saving")

        if not path.endswith((".pt", ".pth")):
            raise ValueError("Save path should end with '.pt' or '.pth'")

        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)

        model_name = os.path.splitext(os.path.basename(path))[0]
        save_dict = {
            "model_state_dict": self.model.state_dict(),
            "hyperparameters": self.get_params(),
            "fitting_epoch": self.fitting_epoch,
            "fitting_loss": self.fitting_loss,
            "fitting_loss_mean": self.fitting_loss_mean,
            "model_name": model_name,
            "date_saved": datetime.datetime.now().isoformat(),
            "version": getattr(self, "__version__", "1.0.0"),
        }
        torch.save(save_dict, path)

    def load_model(self, path: str, repo: Optional[str] = None) -> "MolecularGraphPredictor":
        """Load a saved model from disk or download from HuggingFace hub.

        Parameters
        ----------
        path : str
            Path to the saved model file or desired local path for downloaded model
        repo : str, optional
            HuggingFace model repository ID (e.g., 'username/model-name')
            If provided and local path doesn't exist, will attempt to download from hub

        Returns
        -------
        self : MolecularGraphPredictor
            Updated model instance with loaded weights and parameters

        Raises
        ------
        FileNotFoundError
            If the model file doesn't exist locally and no repo is provided,
            or if download from repo fails
        ValueError
            If the saved file is corrupted or incompatible
        """
        # First try to load from local path
        if os.path.exists(path):
            try:
                checkpoint = torch.load(path, map_location=self.device)
            except Exception as e:
                raise ValueError(f"Error loading model from {path}: {str(e)}")

        # If local file doesn't exist and repo is provided, try downloading
        elif repo is not None:
            try:
                from huggingface_hub import hf_hub_download
                # Create directory if it doesn't exist
                os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)

                # Download from HuggingFace using specified parameters
                model_name = os.path.splitext(os.path.basename(path))[0]
                downloaded_path = hf_hub_download(
                    repo_id=repo,
                    filename=model_name,
                    local_dir=os.path.dirname(path),
                    local_dir_use_symlinks=False,
                )

                # Load the downloaded model
                checkpoint = torch.load(downloaded_path, map_location=self.device)

            except Exception as e:
                if os.path.exists(path):
                    os.remove(path)  # Clean up partial downloads
                raise FileNotFoundError(
                    f"Failed to download model from repository '{repo}': {str(e)}"
                )

        # If neither local file exists nor repo provided
        else:
            raise FileNotFoundError(f"No model file found at '{path}' and no repository provided")

        try:
            # Validate checkpoint contents
            required_keys = {
                "model_state_dict",
                "hyperparameters",
                "fitting_loss",
                "fitting_epoch",
                "model_name",
            }
            if not all(key in checkpoint for key in required_keys):
                missing_keys = required_keys - set(checkpoint.keys())
                raise ValueError(f"Checkpoint missing required keys: {missing_keys}")

            # Validate compatibility of hyperparameters
            for key, value in checkpoint["hyperparameters"].items():
                if hasattr(self, key) and getattr(self, key) != value:
                    warnings.warn(
                        f"Loaded model has different {key} ({value}) "
                        f"than current instance ({getattr(self, key)}). "
                        "Updating to match loaded model."
                    )
                    setattr(self, key, value)

            # Update model state
            if not hasattr(self, 'model') or self.model is None:
                raise ValueError("Model not initialized. Please initialize model before loading weights.")
            
            try:
                self.model.load_state_dict(checkpoint["model_state_dict"])
            except Exception as e:
                raise ValueError(f"Failed to load model state dict: {str(e)}")

            # Update training state
            self.fitting_epoch = checkpoint.get("fitting_epoch", 0)
            self.fitting_loss = checkpoint.get("fitting_loss", float("inf"))
            self.model_name = checkpoint['model_name']
            self.is_fitted_ = True

            # Move model to correct device
            self.model = self.model.to(self.device)

            print(f"Model successfully loaded from {'repository' if repo else 'local path'}")
            return self

        except Exception as e:
            raise ValueError(f"Error loading model checkpoint: {str(e)}")

        finally:
            # Clean up any temporary files
            if repo is not None and os.path.exists(downloaded_path) and downloaded_path != path:
                try:
                    os.remove(downloaded_path)
                except Exception:
                    pass



from enum import Enum
from typing import Dict, Any, Union, List, Tuple, NamedTuple

class ParameterType(Enum):
    """Enum defining types of hyperparameters for optimization.

    Each type corresponds to a specific Optuna suggest method and parameter behavior.
    """

    CATEGORICAL = "categorical"  # Uses suggest_categorical for discrete choices
    INTEGER = "integer"  # Uses suggest_int for whole numbers
    FLOAT = "float"  # Uses suggest_float for continuous values
    LOG_FLOAT = "log_float"  # Uses suggest_float with log=True for exponential scale

class ParameterSpec(NamedTuple):
    """Specification for a hyperparameter including its type and valid range/options."""

    param_type: ParameterType
    value_range: Union[Tuple[Any, Any], List[Any]]

def suggest_parameter(trial: Any, param_name: str, param_spec: ParameterSpec) -> Any:
    """Suggest a parameter value using the appropriate Optuna suggest method.

    Parameters
    ----------
    trial : optuna.Trial
        The Optuna trial object
    param_name : str
        Name of the parameter
    param_spec : ParameterSpec
        Specification of the parameter type and range

    Returns
    -------
    Any
        The suggested parameter value

    Raises
    ------
    ValueError
        If the parameter type is not recognized
    """
    if param_spec.param_type == ParameterType.CATEGORICAL:
        return trial.suggest_categorical(param_name, param_spec.value_range)

    elif param_spec.param_type == ParameterType.INTEGER:
        min_val, max_val = param_spec.value_range
        return trial.suggest_int(param_name, min_val, max_val)

    elif param_spec.param_type == ParameterType.FLOAT:
        min_val, max_val = param_spec.value_range
        return trial.suggest_float(param_name, min_val, max_val)

    elif param_spec.param_type == ParameterType.LOG_FLOAT:
        min_val, max_val = param_spec.value_range
        return trial.suggest_float(param_name, min_val, max_val, log=True)

    else:
        raise ValueError(f"Unknown parameter type: {param_spec.param_type}")
    
def parse_list_params(params_str):
    if params_str is None:
        return None
    return params_str.split(',')


import os
import numpy as np
import warnings
import datetime
from tqdm import tqdm
from typing import Optional, Union, Dict, Any, List, Type
from dataclasses import dataclass, field

import torch
from torch_geometric.data import Data

from .model import GRIN
from ...utils import graph_from_smiles
from ..gnn.modeling_gnn import GNNMolecularPredictor
from ...utils.search import (
    ParameterSpec,
    ParameterType,
)

@dataclass
class GRINMolecularPredictor(GNNMolecularPredictor):
    """
    This predictor implements GRIN for Max Spanning Tree algorithm aligned GNN.

    The full name of GRIN is Graph Invariant Representation Learning.

    References
    ----------
    - Learning Repetition-Invariant Representations for Polymer Informatics.
      https://arxiv.org/pdf/2505.10726

    :param l1_penalty: Weight for the L1 penalty
    :type l1_penalty: float, default=1e-3
    :param epochs_to_penalize: Number of epochs to train before starting L1 penalty
    :type epochs_to_penalize: int, default=100
    """

    l1_penalty: float = 1e-3
    epochs_to_penalize: int = 100

    # Other Non-init fields
    model_name: str = "GRINMolecularPredictor"
    model_class: Type[GRIN] = field(default=GRIN, init=False)
    
    def __post_init__(self):
        super().__post_init__()
    
    @staticmethod
    def _get_param_names() -> List[str]:
        return GNNMolecularPredictor._get_param_names() + [
            "l1_penalty",
            "epochs_to_penalize"
        ]
    def _get_default_search_space(self):
        search_space = super()._get_default_search_space().copy()
        search_space["l1_penalty"] = ParameterSpec(ParameterType.LOG_FLOAT, (1e-6, 1))
        search_space["epochs_to_penalize"] = ParameterSpec(ParameterType.INTEGER, (0, 100))
        return search_space

    def _get_model_params(self, checkpoint: Optional[Dict] = None) -> Dict[str, Any]:
        base_params = super()._get_model_params(checkpoint)
        return base_params

    def _train_epoch(self, train_loader, optimizer, epoch):
        self.model.train()
        losses = []

        iterator = (
            tqdm(train_loader, desc="Training", leave=False)
            if self.verbose
            else train_loader
        )

        for step, batch in enumerate(iterator):
            batch = batch.to(self.device)
            optimizer.zero_grad()
            if epoch >= self.epochs_to_penalize:
                l1_penalty = min(epoch - self.epochs_to_penalize, 1) * self.l1_penalty
            else:
                l1_penalty = 0
            loss = self.model.compute_loss(batch, self.loss_criterion, l1_penalty)
            loss.backward()
            if self.grad_clip_value is not None:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip_value)
            optimizer.step()

            if self.verbose:
                iterator.set_postfix({"Epoch": epoch, "Total Loss": f"{loss.item():.4f}"})
            losses.append(loss.item())

        return losses


import numpy as np
from tqdm import tqdm
from typing import Optional, Union, Dict, Any, Tuple, List, Type
from dataclasses import dataclass, field

import torch
import torch.nn.functional as F
from torch_geometric.loader import DataLoader
from torch_geometric.data import Data

from .sde import load_sde
from .solver import get_pc_sampler
from .model import GDSSModel, get_sde_loss_fn
from .utils import compute_dataset_info, to_dense, quantize_mol

from ...base import BaseMolecularGenerator
from ...utils import graph_from_smiles, graph_to_smiles

@dataclass
class GDSSMolecularGenerator(BaseMolecularGenerator):
    """ This generator implements "Score-based Generative Modeling of Graphs via the System of Stochastic Differential Equations"

    References
    ----------
    - Paper: https://arxiv.org/abs/2202.02514
    - Official Implementation: https://github.com/harryjo97/GDSS

    
    Parameters
    ----------
    num_layer : int, default=3
        Number of layers in the score networks.
    hidden_size_adj : float, default=8
        Hidden dimension size for the adjacency in the adjacency score network.
    hidden_size : int, default=16
        Hidden dimension size latent representation.
    attention_dim : int, default=16
        Dimension of attention layers.
    num_head : int, default=4
        Number of attention heads.
    sde_type_x : str, default='VE'
        SDE type for node features. One of 'VP', 'VE', 'subVP'.
    sde_beta_min_x : float, default=0.1
        Minimum noise level for node features.
    sde_beta_max_x : float, default=1
        Maximum noise level for node features.
    sde_num_scales_x : int, default=1000
        Number of noise scales for node features.
    sde_type_adj : str, default='VE'
        SDE type for adjacency matrix. One of 'VP', 'VE', 'subVP'.
    sde_beta_min_adj : float, default=0.1
        Minimum noise level for adjacency matrix.
    sde_beta_max_adj : float, default=1
        Maximum noise level for adjacency matrix.
    sde_num_scales_adj : int, default=1000
        Number of noise scales for adjacency matrix.
    batch_size : int, default=128
        Batch size for training.
    epochs : int, default=500
        Number of training epochs.
    learning_rate : float, default=0.005
        Learning rate for optimizer.
    grad_clip_value : Optional[float], default=1
        Value for gradient clipping. None means no clipping.
    weight_decay : float, default=1e-4
        Weight decay for optimizer.
    use_loss_reduce_mean : bool, default=False
        Whether to use mean reduction for loss calculation.
    use_lr_scheduler : bool, default=False
        Whether to use learning rate scheduler.
    scheduler_factor : float, default=0.5
        Factor by which to reduce learning rate when using scheduler (only used if use_lr_scheduler is True).
    scheduler_patience : int, default=5
        Number of epochs with no improvement after which learning rate will be reduced (only used if use_lr_scheduler is True).
    sampler_predictor : str, default='Reverse'
        Predictor method for sampling. One of 'Euler', 'Reverse'.
    sampler_corrector : str, default='Langevin'
        Corrector method for sampling. One of 'Langevin', 'None'.
    sampler_snr : float, default=0.2
        Signal-to-noise ratio for corrector.
    sampler_scale_eps : float, default=0.7
        Scale factor for noise level in corrector.
    sampler_n_steps : int, default=1
        Number of corrector steps per predictor step.
    sampler_probability_flow : bool, default=False
        Whether to use probability flow ODE for sampling.
    sampler_noise_removal : bool, default=True
        Whether to remove noise in the final step of sampling.
    verbose : bool, default=False
        Whether to display progress bars and logs.
        
    """
    
    # Model parameters
    num_layer: int = 3
    hidden_size_adj: float = 8
    hidden_size: int = 16
    attention_dim: int = 16
    num_head: int = 4

    # Diffusion parameters
    sde_type_x: str = 'VE' # One of 'VP', 'VE', 'subVP'
    sde_beta_min_x: float = 0.1
    sde_beta_max_x: float = 1
    sde_num_scales_x: int = 1000
    sde_type_adj: str = 'VE' # One of 'VP', 'VE', 'subVP'
    sde_beta_min_adj: float = 0.1
    sde_beta_max_adj: float = 1
    sde_num_scales_adj: int = 1000

    # Training parameters
    batch_size: int = 128
    epochs: int = 500
    learning_rate: float = 0.005
    grad_clip_value: Optional[float] = 1.0
    weight_decay: float = 1e-4
    use_loss_reduce_mean: bool = False

    # Scheduler parameters
    use_lr_scheduler: bool = False
    scheduler_factor: float = 0.5
    scheduler_patience: int = 5

    # Sampling parameters
    sampler_predictor: str = 'Reverse'  # One of 'Euler', 'Reverse', others will be treated as 'Euler'
    sampler_corrector: str = 'Langevin' # One of 'Langevin', 'None', others will be treated as 'None'
    sampler_snr: float = 0.2
    sampler_scale_eps: float = 0.7
    sampler_n_steps: int = 1
    sampler_probability_flow: bool = False
    sampler_noise_removal: bool = True

    verbose: bool = False
    # Attributes
    model_name: str = "GDSSMolecularGenerator"
    fitting_loss: List[float] = field(default_factory=list, init=False)
    fitting_epoch: int = field(default=0, init=False)
    model_class: Type[GDSSModel] = field(default=GDSSModel, init=False)
    # dataset_info: Dict[str, Any] = field(default_factory=dict, init=False)
    def __post_init__(self):
        """Initialize the model after dataclass initialization."""
        super().__post_init__()
        self.max_node = None
        self.input_dim_X = None
        self.input_dim_adj = None
        self.conv = 'GCN'
        self.sampler_eps = 1e-4
        self.train_eps = 1e-5
        self.dataset_info = None
        self.loss_fn = None
        self.sampling_fn = None

    @staticmethod
    def _get_param_names() -> List[str]:
        """Get parameter names for the estimator.

        Returns
        -------
        List[str]
            List of parameter names that can be used for model configuration.
        """

        return [
            # Model Hyperparameters
            "max_node", "num_layer", "num_head", "input_dim_X", "input_dim_adj", "hidden_size_adj", "hidden_size", "attention_dim",
            # Diffusion parameters  
            "dataset_info", "sde_type_x", "sde_beta_min_x", "sde_beta_max_x", "sde_num_scales_x", 
            "sde_type_adj", "sde_beta_min_adj", "sde_beta_max_adj", "sde_num_scales_adj",
            # Training Parameters
            "batch_size", "epochs", "learning_rate", "grad_clip_value", 
            "weight_decay", "use_loss_reduce_mean", "train_eps", "conv",
            # Scheduler Parameters
            "use_lr_scheduler", "scheduler_factor", "scheduler_patience",
            # Sampling Parameters
            "sampler_predictor", "sampler_corrector", "sampler_snr", "sampler_scale_eps", 
            "sampler_n_steps", "sampler_probability_flow", "sampler_noise_removal", "sampler_eps",
            # Other Parameters
            "fitting_epoch", "fitting_loss", "device", "verbose", "model_name"
        ]
    
    def _get_model_params(self, checkpoint: Optional[Dict] = None) -> Dict[str, Any]:
        params = ["input_dim_X", "max_node", "hidden_size", "num_layer", "input_dim_adj", "hidden_size_adj", "attention_dim", "num_head"]

        if checkpoint is not None:
            if "hyperparameters" not in checkpoint:
                raise ValueError("Checkpoint missing 'hyperparameters' key")
            return {k: checkpoint["hyperparameters"][k] for k in params}
        
        return {k: getattr(self, k) for k in params}
        
    def _convert_to_pytorch_data(self, X, y=None):
        """Convert numpy arrays to PyTorch Geometric data format.
        """
        if self.verbose:
            iterator = tqdm(enumerate(X), desc="Converting molecules to graphs", total=len(X))
        else:
            iterator = enumerate(X)

        pyg_graph_list = []
        for idx, smiles_or_mol in iterator:
            if y is not None:
                properties = y[idx]
            else: 
                properties = None
            graph = graph_from_smiles(smiles_or_mol, properties)
            g = Data()
            
            # No H, first heavy atom has type 0
            node_type = torch.from_numpy(graph['node_feat'][:, 0] - 1)
            
            # Filter out invalid node types (< 0)
            valid_mask = node_type >= 0
            if not valid_mask.all():
                # Get valid nodes and adjust edge indices
                valid_indices = torch.where(valid_mask)[0]
                index_map = -torch.ones(node_type.size(0), dtype=torch.long)
                index_map[valid_indices] = torch.arange(valid_indices.size(0))
                
                # Filter edges that connect to invalid nodes
                edge_index = torch.from_numpy(graph["edge_index"])
                valid_edges_mask = valid_mask[edge_index[0]] & valid_mask[edge_index[1]]
                valid_edge_index = edge_index[:, valid_edges_mask]
                
                # Remap edge indices to account for removed nodes
                remapped_edge_index = index_map[valid_edge_index]
                
                # Filter edge attributes
                edge_attr = torch.from_numpy(graph["edge_feat"])[:, 0] + 1
                valid_edge_attr = edge_attr[valid_edges_mask]
                
                # Update node and edge data
                node_type = node_type[valid_mask]
                g.edge_index = remapped_edge_index
                g.edge_attr = valid_edge_attr.long().squeeze(-1)
            else:
                # No invalid nodes, proceed normally
                g.edge_index = torch.from_numpy(graph["edge_index"])
                edge_attr = torch.from_numpy(graph["edge_feat"])[:, 0] + 1
                g.edge_attr = edge_attr.long().squeeze(-1)
            
            # * is encoded as "misc" which is 119 - 1 and should be 117
            node_type[node_type == 118] = 117
            g.x = node_type.long().squeeze(-1)
            del graph["node_feat"]
            del graph["edge_index"]
            del graph["edge_feat"]

            g.y = torch.from_numpy(graph["y"])
            del graph["y"]

            pyg_graph_list.append(g)

        return pyg_graph_list

    def _setup_diffusion_params(self, X: Union[List, Dict]) -> None:
        # Extract dataset info from X if it's a dict (from checkpoint), otherwise compute it
        if isinstance(X, dict):
            dataset_info = X["hyperparameters"]["dataset_info"]
            max_node = X["hyperparameters"]["max_node"]
        else:
            assert isinstance(X, list)
            dataset_info = compute_dataset_info(X)
            max_node = dataset_info["max_node"]

        self.input_dim_X = dataset_info["x_margins"].shape[0]
        self.input_dim_adj = dataset_info["e_margins"].shape[0]
        self.dataset_info = dataset_info
        self.max_node = max_node

        x_sde = load_sde(self.sde_type_x, self.sde_beta_min_x, self.sde_beta_max_x, self.sde_num_scales_x)
        adj_sde = load_sde(self.sde_type_adj, self.sde_beta_min_adj, self.sde_beta_max_adj, self.sde_num_scales_adj)

        self.loss_fn = get_sde_loss_fn(
            x_sde,
            adj_sde,
            train=True,
            reduce_mean=self.use_loss_reduce_mean,
            continuous=True,
            likelihood_weighting=False,
            eps=self.train_eps,
        )

        self.sampling_fn = get_pc_sampler(
            sde_x=x_sde,
            sde_adj=adj_sde,
            predictor=self.sampler_predictor,
            corrector=self.sampler_corrector,
            snr=self.sampler_snr,
            scale_eps=self.sampler_scale_eps,
            n_steps=self.sampler_n_steps,
            probability_flow=self.sampler_probability_flow,
            continuous=True,
            denoise=self.sampler_noise_removal,
            eps=self.sampler_eps,
            device=self.device,
        )


    def _initialize_model(
        self,
        model_class: Type[torch.nn.Module],
        checkpoint: Optional[Dict] = None
    ) -> torch.nn.Module:
        """Initialize the model with parameters or a checkpoint."""
        model_params = self._get_model_params(checkpoint)
        self.model = model_class(**model_params)
        self.model = self.model.to(self.device)
        
        if checkpoint is not None:
            self._setup_diffusion_params(checkpoint)
            self.model.load_state_dict(checkpoint["model_state_dict"])
        return self.model

    def _setup_optimizers(self) -> Tuple[torch.optim.Optimizer, Optional[Any]]:
        """Setup optimization components including optimizer and learning rate scheduler.
        """
        optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay
        )
        scheduler = None
        if self.use_lr_scheduler:
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                optimizer,
                mode="min",
                factor=self.scheduler_factor,
                patience=self.scheduler_patience,
                min_lr=1e-6,
                cooldown=0,
                eps=1e-8,
            )

        return optimizer, scheduler

    def fit(
        self,
        X_train: List[str],
    ) -> "GDSSMolecularGenerator":
        """Fit the model to the training data.

        Parameters
        ----------
        X_train : List[str]
            List of training data in SMILES format.

        Returns
        -------
        self : GDSSMolecularGenerator
            The fitted model.
        """
        X_train, _ = self._validate_inputs(X_train)
        self._setup_diffusion_params(X_train)
        self._initialize_model(self.model_class)
        self.model.initialize_parameters()

        optimizer, scheduler = self._setup_optimizers()
        train_dataset = self._convert_to_pytorch_data(X_train)
        train_loader = DataLoader(
            train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=0
        )

        self.fitting_loss = []
        self.fitting_epoch = 0
        for epoch in range(self.epochs):
            train_losses = self._train_epoch(train_loader, optimizer, epoch)
            self.fitting_loss.append(np.mean(train_losses).item())
            if scheduler:
                scheduler.step(np.mean(train_losses).item())

        self.fitting_epoch = epoch
        self.is_fitted_ = True
        return self
    
    def _train_epoch(self, train_loader, optimizer, epoch):
        self.model.train()
        losses = []
        iterator = (
            tqdm(train_loader, desc="Training", leave=False)
            if self.verbose
            else train_loader
        )
        active_index = self.dataset_info["active_index"]
        for step, batched_data in enumerate(iterator):
            batched_data = batched_data.to(self.device)
            optimizer.zero_grad()

            data_x = F.one_hot(batched_data.x, num_classes=118).float()[:, active_index]
            data_edge_attr = batched_data.edge_attr.float()
            X, E, node_mask = to_dense(data_x, batched_data.edge_index, data_edge_attr, batched_data.batch, self.max_node)
            
            loss_x, loss_adj = self.model.compute_loss(x=X, adj=E, flags=node_mask, loss_fn=self.loss_fn)
            loss = loss_x + loss_adj
            loss.backward()
            if self.grad_clip_value is not None:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip_value)

            optimizer.step()
            losses.append(loss.item())

            if self.verbose:
                iterator.set_postfix({"Epoch": epoch, "Loss": f"{loss.item():.4f}", "Loss_x": f"{loss_x.item():.4f}", "Loss_adj": f"{loss_adj.item():.4f}"})
            
        return losses
    
    @torch.no_grad()
    def generate(self, num_nodes: Optional[Union[List[List], np.ndarray, torch.Tensor]] = None, batch_size: int = 32) -> List[str]:
        """Randomly generate molecules with specified node counts.

        Parameters
        ----------
        num_nodes : Optional[Union[List[List], np.ndarray, torch.Tensor]], default=None
            Number of nodes for each molecule in the batch. If None, samples from
            the training distribution. Can be provided as:
            - A list of lists
            - A numpy array of shape (batch_size, 1) 
            - A torch tensor of shape (batch_size, 1)
            
        batch_size : int, default=32
            Number of molecules to generate.

        Returns
        -------
        List[str]
            List of generated molecules in SMILES format.
        """
        if not self.is_fitted_:
            raise ValueError("Model must be fitted before generating molecules.")

        if num_nodes is not None:
            batch_size = len(num_nodes)

        if num_nodes is None:
            num_nodes_dist = self.dataset_info["num_nodes_dist"]
            num_nodes = num_nodes_dist.sample_n(batch_size, self.device)
        elif isinstance(num_nodes, list):
            num_nodes = torch.tensor(num_nodes).to(self.device)
        elif isinstance(num_nodes, np.ndarray):
            num_nodes = torch.from_numpy(num_nodes).to(self.device)
        if num_nodes.dim() == 1:
            num_nodes = num_nodes.unsqueeze(-1)
        
        assert num_nodes.size(0) == batch_size
        arange = (
            torch.arange(self.max_node).to(self.device)
            .unsqueeze(0)
            .expand(batch_size, -1)
        )
        node_mask = arange < num_nodes

        if not hasattr(self, 'dataset_info') or self.dataset_info is None:
            raise ValueError("Dataset info not found. Please call setup_diffusion_params first.")
        
        shape_x = (
            batch_size,
            self.max_node,
            self.input_dim_X,
        )
        shape_adj = (batch_size, self.max_node, self.max_node)

        X, E, _ = self.sampling_fn(self.model.score_network_x, self.model.score_network_a, shape_x, shape_adj, node_mask)
        E = quantize_mol(E)
        X = X.argmax(dim=-1)

        molecule_list = []
        for i in range(batch_size):
            n = num_nodes[i][0].item()
            atom_types = X[i, :n].cpu()
            edge_types = E[i, :n, :n].cpu()
            molecule_list.append([atom_types, edge_types])

        smiles_list = graph_to_smiles(molecule_list, self.dataset_info["atom_decoder"])
        return smiles_list


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

def create_var(tensor, requires_grad=None):
    if requires_grad is None:
        return Variable(tensor).cuda()
    else:
        return Variable(tensor, requires_grad=requires_grad).cuda()

def index_select_ND(source, dim, index):
    index_size = index.size()
    suffix_dim = source.size()[1:]
    final_size = index_size + suffix_dim
    target = source.index_select(dim, index.view(-1))
    return target.view(final_size)

def avg_pool(all_vecs, scope, dim):
    size = create_var(torch.Tensor([le for _,le in scope]))
    return all_vecs.sum(dim=dim) / size.unsqueeze(-1)

def stack_pad_tensor(tensor_list):
    max_len = max([t.size(0) for t in tensor_list])
    for i,tensor in enumerate(tensor_list):
        pad_len = max_len - tensor.size(0)
        tensor_list[i] = F.pad( tensor, (0,0,0,pad_len) )
    return torch.stack(tensor_list, dim=0)

#3D padded tensor to 2D matrix, with padded zeros removed
def flatten_tensor(tensor, scope):
    assert tensor.size(0) == len(scope)
    tlist = []
    for i,tup in enumerate(scope):
        le = tup[1]
        tlist.append( tensor[i, 0:le] )
    return torch.cat(tlist, dim=0)

#2D matrix to 3D padded tensor
def inflate_tensor(tensor, scope): 
    max_len = max([le for _,le in scope])
    batch_vecs = []
    for st,le in scope:
        cur_vecs = tensor[st : st + le]
        cur_vecs = F.pad( cur_vecs, (0,0,0,max_len-le) )
        batch_vecs.append( cur_vecs )

    return torch.stack(batch_vecs, dim=0)

def GRU(x, h_nei, W_z, W_r, U_r, W_h):
    hidden_size = x.size()[-1]
    sum_h = h_nei.sum(dim=1)
    z_input = torch.cat([x,sum_h], dim=1)
    z = F.sigmoid(W_z(z_input))

    r_1 = W_r(x).view(-1,1,hidden_size)
    r_2 = U_r(h_nei)
    r = F.sigmoid(r_1 + r_2)
    
    gated_h = r * h_nei
    sum_gated_h = gated_h.sum(dim=1)
    h_input = torch.cat([x,sum_gated_h], dim=1)
    pre_h = F.tanh(W_h(h_input))
    new_h = (1.0 - z) * sum_h + z * pre_h
    return new_h




import torch
import torch.nn as nn
import rdkit.Chem as Chem
import torch.nn.functional as F


from .nnutils import *
from .chemutils import get_mol

ELEM_LIST = ['C', 'N', 'O', 'S', 'F', 'Si', 'P', 'Cl', 'Br', 'Mg', 'Na', 'Ca', 'Fe', 'Al', 'I', 'B', 'K', 'Se', 'Zn', 'H', 'Cu', 'Mn', 'unknown']

ATOM_FDIM = len(ELEM_LIST) + 6 + 5 + 4 + 1
BOND_FDIM = 5 + 6
MAX_NB = 6

def onek_encoding_unk(x, allowable_set):
    if x not in allowable_set:
        x = allowable_set[-1]
    return list(map(lambda s: x == s, allowable_set))

def atom_features(atom):
    return torch.Tensor(onek_encoding_unk(atom.GetSymbol(), ELEM_LIST) 
            + onek_encoding_unk(atom.GetDegree(), [0,1,2,3,4,5]) 
            + onek_encoding_unk(atom.GetFormalCharge(), [-1,-2,1,2,0])
            + onek_encoding_unk(int(atom.GetChiralTag()), [0,1,2,3])
            + [atom.GetIsAromatic()])

def bond_features(bond):
    bt = bond.GetBondType()
    stereo = int(bond.GetStereo())
    fbond = [bt == Chem.rdchem.BondType.SINGLE, bt == Chem.rdchem.BondType.DOUBLE, bt == Chem.rdchem.BondType.TRIPLE, bt == Chem.rdchem.BondType.AROMATIC, bond.IsInRing()]
    fstereo = onek_encoding_unk(stereo, [0,1,2,3,4,5])
    return torch.Tensor(fbond + fstereo)

class MPN(nn.Module):

    def __init__(self, hidden_size, depth):
        super(MPN, self).__init__()
        self.hidden_size = hidden_size
        self.depth = depth

        self.W_i = nn.Linear(ATOM_FDIM + BOND_FDIM, hidden_size, bias=False)
        self.W_h = nn.Linear(hidden_size, hidden_size, bias=False)
        self.W_o = nn.Linear(ATOM_FDIM + hidden_size, hidden_size)

    def forward(self, fatoms, fbonds, agraph, bgraph, scope):
        fatoms = create_var(fatoms)
        fbonds = create_var(fbonds)
        agraph = create_var(agraph)
        bgraph = create_var(bgraph)

        binput = self.W_i(fbonds)
        message = F.relu(binput)

        for i in range(self.depth - 1):
            nei_message = index_select_ND(message, 0, bgraph)
            nei_message = nei_message.sum(dim=1)
            nei_message = self.W_h(nei_message)
            message = F.relu(binput + nei_message)

        nei_message = index_select_ND(message, 0, agraph)
        nei_message = nei_message.sum(dim=1)
        ainput = torch.cat([fatoms, nei_message], dim=1)
        atom_hiddens = F.relu(self.W_o(ainput))

        max_len = max([x for _,x in scope])
        batch_vecs = []
        for st,le in scope:
            cur_vecs = atom_hiddens[st : st + le].mean(dim=0)
            batch_vecs.append( cur_vecs )

        mol_vecs = torch.stack(batch_vecs, dim=0)
        return mol_vecs 

    @staticmethod
    def tensorize(mol_batch):
        padding = torch.zeros(ATOM_FDIM + BOND_FDIM)
        fatoms,fbonds = [],[padding] #Ensure bond is 1-indexed
        in_bonds,all_bonds = [],[(-1,-1)] #Ensure bond is 1-indexed
        scope = []
        total_atoms = 0

        for smiles in mol_batch:
            mol = get_mol(smiles)
            #mol = Chem.MolFromSmiles(smiles)
            n_atoms = mol.GetNumAtoms()
            for atom in mol.GetAtoms():
                fatoms.append( atom_features(atom) )
                in_bonds.append([])

            for bond in mol.GetBonds():
                a1 = bond.GetBeginAtom()
                a2 = bond.GetEndAtom()
                x = a1.GetIdx() + total_atoms
                y = a2.GetIdx() + total_atoms

                b = len(all_bonds) 
                all_bonds.append((x,y))
                fbonds.append( torch.cat([fatoms[x], bond_features(bond)], 0) )
                in_bonds[y].append(b)

                b = len(all_bonds)
                all_bonds.append((y,x))
                fbonds.append( torch.cat([fatoms[y], bond_features(bond)], 0) )
                in_bonds[x].append(b)
            
            scope.append((total_atoms,n_atoms))
            total_atoms += n_atoms

        total_bonds = len(all_bonds)
        fatoms = torch.stack(fatoms, 0)
        fbonds = torch.stack(fbonds, 0)
        agraph = torch.zeros(total_atoms,MAX_NB).long()
        bgraph = torch.zeros(total_bonds,MAX_NB).long()

        for a in range(total_atoms):
            for i,b in enumerate(in_bonds[a]):
                if i >= MAX_NB: break
                agraph[a,i] = b

        for b1 in range(1, total_bonds):
            x,y = all_bonds[b1]
            for i,b2 in enumerate(in_bonds[x]):
                if all_bonds[b2][0] != y:
                    if i >= MAX_NB: break
                    bgraph[b1,i] = b2

        return (fatoms, fbonds, agraph, bgraph, scope)




import torch.nn as nn

class MLP(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        bias=True,
        drop=0.,
        use_bn=True,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        linear_layer = nn.Linear

        self.fc1 = linear_layer(in_features, hidden_features, bias=bias)
        if use_bn:
            self.bn1 = nn.BatchNorm1d(hidden_features)
        else:
            self.bn1 = nn.Identity()
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop)
        self.fc2 = linear_layer(hidden_features, out_features, bias=bias)

    def forward(self, x):
        x = self.fc1(x)
        x = self.bn1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.fc2(x)
        return x


import torch
import torch.nn as nn
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool

from ...nn import GNN_node, GNN_node_Virtualnode, MLP
from ...utils import init_weights

reg_criterion = torch.nn.L1Loss()
class_criterion = torch.nn.BCEWithLogitsLoss()

class GNN(nn.Module):
    def __init__(
        self,
        num_layer,
        hidden_size,
        num_task,
        drop_ratio=0.5,
        norm_layer="batch_norm",
        encoder_type="gin-virtual",
        readout="max",
    ):
        super(GNN, self).__init__()
        gnn_name = encoder_type.split("-")[0]
        self.num_task = num_task
        self.hidden_size = hidden_size

        encoder_params = {
            "num_layer": num_layer,
            "hidden_size": hidden_size,
            "JK": "last", 
            "drop_ratio": drop_ratio,
            "residual": True,
            "gnn_name": gnn_name,
            "norm_layer": norm_layer
        }
        
        # Choose encoder type based on encoder
        encoder_class = GNN_node_Virtualnode if "virtual" in encoder_type else GNN_node
        self.graph_encoder = encoder_class(**encoder_params)
        pooling_funcs = {
            "sum": global_add_pool,
            "mean": global_mean_pool,
            "max": global_max_pool
        }
        self.readout = pooling_funcs.get(readout)
        if self.readout is None:
            raise ValueError(f"Invalid graph pooling type {readout}.")

        self.predictor = MLP(hidden_size, hidden_features=2 * hidden_size, out_features=num_task)
    
    def initialize_parameters(self, seed=None):
        """
        Randomly initialize all model parameters using the init_weights function.
        
        Args:
            seed (int, optional): Random seed for reproducibility. Defaults to None.
        """
        if seed is not None:
            torch.manual_seed(seed)
        
        # Initialize the main components
        init_weights(self.graph_encoder)
        init_weights(self.predictor)
        
        # Reset all parameters using PyTorch Geometric's reset function
        def reset_parameters(module):
            if hasattr(module, 'reset_parameters'):
                module.reset_parameters()
            elif hasattr(module, 'weight') and hasattr(module.weight, 'data'):
                init_weights(module)
        
        self.apply(reset_parameters)

    def compute_loss(self, batched_data, is_class):
        h_node, _ = self.graph_encoder(batched_data)
        h_rep = self.readout(h_node, batched_data.batch)
        prediction = self.predictor(h_rep)

        target = batched_data.y.to(torch.float32)
        is_labeled = ~torch.isnan(target)
        is_labeled_class = is_labeled[:, is_class]
        is_labeled_reg = is_labeled[:, ~is_class]

        target_class = target[:, is_class]
        target_reg = target[:, ~is_class]
        prediction_class = prediction[:, is_class]
        prediction_reg = prediction[:, ~is_class]
        
        loss_class = class_criterion(prediction_class.to(torch.float32)[is_labeled_class], target_class[is_labeled_class])
        loss_reg = reg_criterion(prediction_reg.to(torch.float32)[is_labeled_reg], target_reg[is_labeled_reg])
        loss = loss_class + loss_reg
        return loss

    def forward(self, batched_data):
        h_node, _ = self.graph_encoder(batched_data)
        h_rep = self.readout(h_node, batched_data.batch)
        return {"graph": h_rep, "node": h_node}


import random
import copy
import torch
import torch.nn.functional as F
from torch_geometric.data import Data

class DataLoaderMaskingPred(torch.utils.data.DataLoader):
    r"""Data loader which merges data objects from a
    :class:`torch_geometric.data.dataset` to a mini-batch.
    Args:
        dataset (Dataset): The dataset from which to load the data.
        batch_size (int, optional): How may samples per batch to load.
            (default: :obj:`1`)
        shuffle (bool, optional): If set to :obj:`True`, the data will be
            reshuffled at every epoch (default: :obj:`True`)
    """

    def __init__(self, dataset, batch_size=1, shuffle=True, mask_rate=0.0, mask_edge=False, **kwargs):
        self._transform = MaskAtom(num_atom_type = 119, num_edge_type = 5, mask_rate = mask_rate, mask_edge=mask_edge)
        super(DataLoaderMaskingPred, self).__init__(
            dataset,
            batch_size,
            shuffle,
            collate_fn=self.collate_fn,
            **kwargs)
    
    def collate_fn(self, batches):
        batchs = [self._transform(x) for x in batches]
        return BatchMasking.from_data_list(batchs)

class MaskAtom:
    def __init__(self, num_atom_type, num_edge_type, mask_rate, mask_edge=True):
        """
        Randomly masks an atom, and optionally masks edges connecting to it.
        The mask atom type index is num_possible_atom_type
        The mask edge type index in num_possible_edge_type
        :param num_atom_type:
        :param num_edge_type:
        :param mask_rate: % of atoms to be masked
        :param mask_edge: If True, also mask the edges that connect to the
        masked atoms
        """
        self.num_atom_type = num_atom_type
        self.num_edge_type = num_edge_type
        self.mask_rate = mask_rate
        self.mask_edge = mask_edge
        # self.num_chirality_tag = 3
        # self.num_bond_direction = 3 

    def __call__(self, input_data, masked_atom_indices=None):
        """

        :param data: pytorch geometric data object. Assume that the edge
        ordering is the default pytorch geometric ordering, where the two
        directions of a single edge occur in pairs.
        Eg. data.edge_index = tensor([[0, 1, 1, 2, 2, 3],
                                     [1, 0, 2, 1, 3, 2]])
        :param masked_atom_indices: If None, then randomly samples num_atoms
        * mask rate number of atom indices
        Otherwise a list of atom idx that sets the atoms to be masked (for
        debugging only)
        :return: None, Creates new attributes in original data object:
        data.mask_node_idx
        data.mask_node_label
        data.mask_edge_idx
        data.mask_edge_label
        """
        data = copy.deepcopy(input_data)
        if masked_atom_indices == None:
            # sample x distinct atoms to be masked, based on mask rate. But
            # will sample at least 1 atom
            num_atoms = data.x.size()[0]
            sample_size = int(num_atoms * self.mask_rate + 1)
            masked_atom_indices = random.sample(range(num_atoms), sample_size)

        # create mask node label by copying atom feature of mask atom
        mask_node_labels_list = []
        for atom_idx in masked_atom_indices:
            mask_node_labels_list.append(data.x[atom_idx].view(1, -1))
        data.mask_node_label = torch.cat(mask_node_labels_list, dim=0)
        data.masked_atom_indices = torch.tensor(masked_atom_indices)

        # ----------- graphMAE -----------
        atom_type = F.one_hot(data.mask_node_label[:, 0], num_classes=self.num_atom_type).float()
        # atom_chirality = F.one_hot(data.mask_node_label[:, 1], num_classes=self.num_chirality_tag).float()
        # data.node_attr_label = torch.cat((atom_type,atom_chirality), dim=1)
        data.node_attr_label = atom_type

        # modify the original node feature of the masked node
        for atom_idx in masked_atom_indices:
            feat_dim = data.x[atom_idx].shape[0]
            data.x[atom_idx] = torch.zeros(feat_dim)
            data.x[atom_idx][0] = self.num_atom_type - 1

        if self.mask_edge:
            # create mask edge labels by copying edge features of edges that are bonded to
            # mask atoms
            connected_edge_indices = []
            for bond_idx, (u, v) in enumerate(data.edge_index.cpu().numpy().T):
                for atom_idx in masked_atom_indices:
                    if atom_idx in set((u, v)) and \
                        bond_idx not in connected_edge_indices:
                        connected_edge_indices.append(bond_idx)

            if len(connected_edge_indices) > 0:
                # create mask edge labels by copying bond features of the bonds connected to
                # the mask atoms
                mask_edge_labels_list = []
                for bond_idx in connected_edge_indices[::2]: # because the
                    # edge ordering is such that two directions of a single
                    # edge occur in pairs, so to get the unique undirected
                    # edge indices, we take every 2nd edge index from list
                    mask_edge_labels_list.append(
                        data.edge_attr[bond_idx].view(1, -1))

                data.mask_edge_label = torch.cat(mask_edge_labels_list, dim=0)
                # modify the original bond features of the bonds connected to the mask atoms
                for bond_idx in connected_edge_indices:
                    feat_dim = data.edge_attr[bond_idx].shape[0]
                    data.edge_attr[bond_idx] = torch.zeros(feat_dim)
                    data.edge_attr[bond_idx][0] = self.num_edge_type - 1
                    # print('feat_dim', feat_dim, 'data.edge_attr[bond_idx].shape', data.edge_attr[bond_idx].shape)

                data.connected_edge_indices = torch.tensor(
                    connected_edge_indices[::2])
            else:
                data.mask_edge_label = torch.empty((0, 2)).to(torch.int64)
                data.connected_edge_indices = torch.tensor(
                    connected_edge_indices).to(torch.int64)

            edge_type = F.one_hot(data.mask_edge_label[:, 0], num_classes=self.num_edge_type).float()
            # bond_direction = F.one_hot(data.mask_edge_label[:, 1], num_classes=self.num_bond_direction).float()
            data.edge_attr_label = edge_type
            # data.edge_attr_label = edge_type

        return data

    def __repr__(self):
        return '{}(num_atom_type={}, num_edge_type={}, mask_rate={}, mask_edge={})'.format(
            self.__class__.__name__, self.num_atom_type, self.num_edge_type,
            self.mask_rate, self.mask_edge)

class BatchMasking(Data):
    r"""A plain old python object modeling a batch of graphs as one big
    (dicconnected) graph. With :class:`torch_geometric.data.Data` being the
    base class, all its methods can also be used here.
    In addition, single graphs can be reconstructed via the assignment vector
    :obj:`batch`, which maps each node to its respective graph identifier.
    """

    def __init__(self, batch=None, **kwargs):
        super(BatchMasking, self).__init__(**kwargs)
        self.batch = batch

    @staticmethod
    def from_data_list(data_list):
        r"""Constructs a batch object from a python list holding
        :class:`torch_geometric.data.Data` objects.
        The assignment vector :obj:`batch` is created on the fly."""
        keys = [set(data.keys()) for data in data_list]
        keys = list(set.union(*keys))
        assert 'batch' not in keys

        batch = BatchMasking()

        for key in keys:
            batch[key] = []
        batch.batch = []

        cumsum_node = 0
        cumsum_edge = 0

        for i, data in enumerate(data_list):
            num_nodes = data.num_nodes
            batch.batch.append(torch.full((num_nodes, ), i, dtype=torch.long))
            for key in data.keys():
                item = data[key]
                if key in ['edge_index', 'masked_atom_indices']:
                    item = item + cumsum_node
                elif key  == 'connected_edge_indices':
                    item = item + cumsum_edge
                batch[key].append(item)

            cumsum_node += num_nodes
            cumsum_edge += data.edge_index.shape[1]

        for key in keys:
            batch[key] = torch.cat(
                batch[key], dim=data_list[0].__cat_dim__(key, batch[key][0]))
        batch.batch = torch.cat(batch.batch, dim=-1)
        return batch.contiguous()

    def cumsum(self, key, item):
        r"""If :obj:`True`, the attribute :obj:`key` with content :obj:`item`
        should be added up cumulatively before concatenated together.
        .. note::
            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        return key in ['edge_index', 'face', 'masked_atom_indices', 'connected_edge_indices']

    @property
    def num_graphs(self):
        """Returns the number of graphs in the batch."""
        return self.batch[-1].item() + 1



from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from typing import Optional, ClassVar, Union, List, Dict, Any, Tuple, Callable, Type, Literal

import torch
import numpy as np
from .base import BaseModel

@dataclass
class BaseMolecularEncoder(BaseModel, ABC):
    """Base class for molecular representation learning."""
    
    model_name: str = field(default="BaseMolecularEncoder")

    @abstractmethod
    def encode(self, X: List[str], return_type: Literal["np", "pt"] = "pt") -> Union[np.ndarray, torch.Tensor]:
        pass
    
    @abstractmethod
    def fit(self, X: List[str], y: Optional[np.ndarray] = None) -> "BaseMolecularEncoder":
        pass

    def _inspect_task_types(self, y: Union[np.ndarray, torch.Tensor], return_type: Literal["pt", "np"] = "pt") -> Union[np.ndarray, torch.Tensor]:
        """Inspect the task types of the target values.

        Parameters
        ----------
        y : Union[np.ndarray, torch.Tensor]
            Target values: 2D array for multiple tasks
        return_type : Literal["pt", "np"], default="pt"
            Return type of the result

        Returns
        -------
        result : Union[np.ndarray, torch.Tensor]
            Result of the task types inspection
        """
        if isinstance(y, np.ndarray):
            y = torch.from_numpy(y)
        # 0/1/nan -> binary classification (True), otherwise -> regression (False)
        result = torch.tensor([torch.all(torch.isnan(y[:, i]) | (y[:, i] == 0) | (y[:, i] == 1)) for i in range(y.shape[1])], dtype=torch.bool)
        if return_type == "np":
            return result.numpy()
        return result


from rdkit import Chem, RDLogger

RDLogger.DisableLog("rdApp.*")

import re
import random
import logging
from rdkit import Chem
from typing import List, Tuple, Optional
random.seed(0)
import torch

bond_dict = [
    None,
    Chem.rdchem.BondType.SINGLE,
    Chem.rdchem.BondType.DOUBLE,
    Chem.rdchem.BondType.TRIPLE,
    Chem.rdchem.BondType.AROMATIC,
]

ATOM_VALENCY = {6: 4, 7: 3, 8: 2, 9: 1, 15: 3, 16: 2, 17: 1, 35: 1, 53: 1}

def graph_to_smiles(molecule_list: List[Tuple], atom_decoder: list) -> List[Optional[str]]:
    smiles_list = []
    for index, graph in enumerate(molecule_list):
        try:
            atom_types, edge_types = graph
            mol_init = build_molecule_with_partial_charges(atom_types, edge_types, atom_decoder)
            
            # Try to correct the molecule with connection=True, then False if needed
            for connection in (True, False):
                mol_conn, _ = correct_mol(mol_init, connection=connection)
                if mol_conn is not None:
                    break
            else:
                mol_conn = mol_init  # Fallback to initial molecule

            # Convert to SMILES
            smiles = mol2smiles(mol_conn)
            if not smiles:
                smiles = Chem.MolToSmiles(mol_conn)

            if smiles:
                mol = get_mol(smiles)
                if mol is not None:
                    # Get the largest fragment
                    mol_frags = Chem.rdmolops.GetMolFrags(mol, asMols=True, sanitizeFrags=False)
                    largest_mol = max(mol_frags, key=lambda m: m.GetNumAtoms())
                    
                    largest_smiles = mol2smiles(largest_mol)
                    if largest_smiles and len(largest_smiles) > 1:
                        smiles_list.append(largest_smiles)
                    else:
                        smiles_list.append(None)
                else:
                    smiles_list.append(None)
            else:
                smiles_list.append(None)

        except Exception as e:
            try:
                # Fallback to RDKit's MolToSmiles if everything else fails
                fallback_smiles = Chem.MolToSmiles(mol_init)
                if fallback_smiles:
                    smiles_list.append(fallback_smiles)
                else:
                    smiles_list.append(None)
            except Exception as e2:
                smiles_list.append(None)

    return smiles_list

def build_molecule_with_partial_charges(atom_types, edge_types, atom_decoder, verbose=False):
    if verbose:
        print("\nbuilding new molecule")

    mol = Chem.RWMol()
    for atom in atom_types:
        a = Chem.Atom(atom_decoder[atom.item()])
        mol.AddAtom(a)
        if verbose:
            print("Atom added: ", atom.item(), atom_decoder[atom.item()])

    edge_types = torch.triu(edge_types)
    all_bonds = torch.nonzero(edge_types)

    for i, bond in enumerate(all_bonds):
        if bond[0].item() != bond[1].item():
            mol.AddBond(
                bond[0].item(),
                bond[1].item(),
                bond_dict[edge_types[bond[0], bond[1]].item()],
            )
            if verbose:
                print(
                    "bond added:",
                    bond[0].item(),
                    bond[1].item(),
                    edge_types[bond[0], bond[1]].item(),
                    bond_dict[edge_types[bond[0], bond[1]].item()],
                )
            # add formal charge to atom: e.g. [O+], [N+], [S+]
            # not support [O-], [N-], [S-], [NH+] etc.
            flag, atomid_valence = check_valency(mol)
            if verbose:
                print("flag, valence", flag, atomid_valence)
            if flag:
                continue
            else:
                if len(atomid_valence) == 2:
                    idx = atomid_valence[0]
                    v = atomid_valence[1]
                    an = mol.GetAtomWithIdx(idx).GetAtomicNum()
                    if verbose:
                        print("atomic num of atom with a large valence", an)
                    if an in (7, 8, 16) and (v - ATOM_VALENCY[an]) == 1:
                        mol.GetAtomWithIdx(idx).SetFormalCharge(1)
                        # print("Formal charge added")
                else:
                    continue
    return mol


def correct_mol(mol, connection=False):
    #####
    no_correct = False
    flag, _ = check_valency(mol)
    if flag:
        no_correct = True

    while True:
        if connection:
            mol_conn = connect_fragments(mol)
            mol = mol_conn
            if mol is None:
                return None, no_correct
        flag, atomid_valence = check_valency(mol)
        if flag:
            break
        else:
            try:
                assert len(atomid_valence) == 2
                idx = atomid_valence[0]
                v = atomid_valence[1]
                queue = []
                check_idx = 0
                for b in mol.GetAtomWithIdx(idx).GetBonds():
                    type = int(b.GetBondType())
                    queue.append(
                        (b.GetIdx(), type, b.GetBeginAtomIdx(), b.GetEndAtomIdx())
                    )
                    if type == 12:
                        check_idx += 1
                queue.sort(key=lambda tup: tup[1], reverse=True)

                if queue[-1][1] == 12:
                    return None, no_correct
                elif len(queue) > 0:
                    start = queue[check_idx][2]
                    end = queue[check_idx][3]
                    t = queue[check_idx][1] - 1
                    mol.RemoveBond(start, end)
                    if t >= 1:
                        mol.AddBond(start, end, bond_dict[t])
            except Exception as e:
                # print(f"An error occurred in correction: {e}")
                return None, no_correct
    return mol, no_correct


def check_valid(smiles):
    mol = get_mol(smiles)
    if mol is None:
        return False
    smiles = mol2smiles(mol)
    if smiles is None:
        return False
    return True


def get_mol(smiles_or_mol):
    """
    Loads SMILES/molecule into RDKit's object
    """
    if isinstance(smiles_or_mol, str):
        if len(smiles_or_mol) == 0:
            return None
        mol = Chem.MolFromSmiles(smiles_or_mol)
        if mol is None:
            return None
        try:
            Chem.SanitizeMol(mol)
        except ValueError:
            return None
        return mol
    return smiles_or_mol


def mol2smiles(mol):
    if mol is None:
        return None
    try:
        Chem.SanitizeMol(mol)
    except ValueError:
        return None
    return Chem.MolToSmiles(mol)


def check_valency(mol):
    try:
        # First attempt to sanitize with specific properties
        Chem.SanitizeMol(mol, sanitizeOps=Chem.SanitizeFlags.SANITIZE_PROPERTIES)
        return True, None
    except ValueError as e:
        e = str(e)
        p = e.find("#")
        e_sub = e[p:]
        atomid_valence = list(map(int, re.findall(r"\d+", e_sub)))
        return False, atomid_valence
    except Exception as e:
        # print(f"An unexpected error occurred: {e}")
        return False, []


##### connect fragements
def select_atom_with_available_valency(frag):
    atoms = list(frag.GetAtoms())
    random.shuffle(atoms)
    for atom in atoms:
        if atom.GetAtomicNum() > 1 and atom.GetImplicitValence() > 0:
            return atom
    return None


def select_atoms_with_available_valency(frag):
    return [
        atom
        for atom in frag.GetAtoms()
        if atom.GetAtomicNum() > 1 and atom.GetImplicitValence() > 0
    ]


def try_to_connect_fragments(combined_mol, frag, atom1, atom2):
    # Make copies of the molecules to try the connection
    trial_combined_mol = Chem.RWMol(combined_mol)
    trial_frag = Chem.RWMol(frag)

    # Add the new fragment to the combined molecule with new indices
    new_indices = {
        atom.GetIdx(): trial_combined_mol.AddAtom(atom)
        for atom in trial_frag.GetAtoms()
    }

    # Add the bond between the suitable atoms from each fragment
    trial_combined_mol.AddBond(
        atom1.GetIdx(), new_indices[atom2.GetIdx()], Chem.BondType.SINGLE
    )

    # Adjust the hydrogen count of the connected atoms
    for atom_idx in [atom1.GetIdx(), new_indices[atom2.GetIdx()]]:
        atom = trial_combined_mol.GetAtomWithIdx(atom_idx)
        num_h = atom.GetTotalNumHs()
        atom.SetNumExplicitHs(max(0, num_h - 1))

    # Add bonds for the new fragment
    for bond in trial_frag.GetBonds():
        trial_combined_mol.AddBond(
            new_indices[bond.GetBeginAtomIdx()],
            new_indices[bond.GetEndAtomIdx()],
            bond.GetBondType(),
        )

    # Convert to a Mol object and try to sanitize it
    new_mol = Chem.Mol(trial_combined_mol)
    try:
        Chem.SanitizeMol(new_mol)
        return new_mol  # Return the new valid molecule
    except Chem.MolSanitizeException:
        return None  # If the molecule is not valid, return None


def connect_fragments(mol):
    # Get the separate fragments
    frags = Chem.GetMolFrags(mol, asMols=True, sanitizeFrags=False)
    if len(frags) < 2:
        return mol

    combined_mol = Chem.RWMol(frags[0])

    for frag in frags[1:]:
        # Select all atoms with available valency from both molecules
        atoms1 = select_atoms_with_available_valency(combined_mol)
        atoms2 = select_atoms_with_available_valency(frag)

        # Try to connect using all combinations of available valency atoms
        for atom1 in atoms1:
            for atom2 in atoms2:
                new_mol = try_to_connect_fragments(combined_mol, frag, atom1, atom2)
                if new_mol is not None:
                    # If a valid connection is made, update the combined molecule and break
                    combined_mol = new_mol
                    break
            else:
                # Continue if the inner loop didn't break (no valid connection found for atom1)
                continue
            # Break if the inner loop did break (valid connection found)
            break
        else:
            # If no valid connections could be made with any of the atoms, return None
            return None

    return combined_mol


import torch
import numpy as np
import types
import inspect

def serialize_config(obj):
    """Helper function to make config JSON serializable.
    
    Handles special cases like lambda functions, torch modules, and numpy arrays.
    
    Parameters
    ----------
    obj : Any
        The object to serialize
        
    Returns
    -------
    Any
        JSON serializable representation of the object
    """
    # Handle None
    if obj is None:
        return None
        
    # Handle lambda and regular functions
    if isinstance(obj, (types.LambdaType, types.FunctionType, types.MethodType)):
        # For lambda functions, try to get the source code
        try:
            if isinstance(obj, types.LambdaType):
                source = inspect.getsource(obj).strip()
                return {"_type": "lambda", "source": source}
            else:
                return {"_type": "function", "name": obj.__name__}
        except (IOError, TypeError):
            # Fallback for built-in functions or when source isn't available
            return {"_type": "function", "name": str(obj)}
    
    # Handle PyTorch modules and optimizers
    elif isinstance(obj, (torch.nn.Module, torch.optim.Optimizer)):
        return {
            "_type": "torch_class",
            "class_name": obj.__class__.__name__,
            "module": obj.__class__.__module__
        }
        
    # Handle PyTorch tensors
    elif isinstance(obj, torch.Tensor):
        # If it's a single number wrapped in a torch tensor, just return the number
        if obj.numel() == 1:
            return obj.item()
        elif obj.numel() < 1000:
            return {
                "_type": "torch_tensor",
                "data": obj.detach().cpu().tolist(),
                "shape": list(obj.shape)
            }
        return {
            "_type": "torch_tensor",
            "shape": list(obj.shape)
        }
        
    # Handle numpy arrays
    elif isinstance(obj, (np.ndarray, np.generic)):
        # If it's a single number wrapped in a numpy array, just return the number
        if obj.size == 1:
            return obj.item()
        elif obj.size < 1000:
            return {
                "_type": "numpy_array",
                "data": obj.tolist(),
                "shape": list(obj.shape)
            }
        return {
            "_type": "numpy_array",
            "shape": list(obj.shape)
        }
        
    # Handle sets and frozensets
    elif isinstance(obj, (set, frozenset)):
        return {
            "_type": "set",
            "data": list(obj)
        }
        
    # Handle custom objects with __dict__
    elif hasattr(obj, '__dict__'):
        return {
            "_type": "custom_class",
            "class_name": obj.__class__.__name__,
            "module": obj.__class__.__module__
        }
        
    # Handle basic types that are JSON serializable
    elif isinstance(obj, (str, int, float, bool)):
        return obj
        
    # Handle any other types by converting to string
    return {
        "_type": "unknown",
        "repr": str(obj)
    }
        

def sanitize_config(config_dict):
    """Recursively sanitize config dictionary for JSON serialization.
    
    Handles nested structures and special cases.
    
    Parameters
    ----------
    config_dict : dict
        Configuration dictionary to sanitize
        
    Returns
    -------
    dict
        Sanitized configuration dictionary that is JSON serializable
    """
    if not isinstance(config_dict, dict):
        return serialize_config(config_dict)
        
    clean_dict = {}
    for key, value in config_dict.items():
        # Skip private attributes and callable objects stored as attributes
        if isinstance(key, str) and (key.startswith('_') or callable(value)):
            continue
            
        # Handle nested dictionaries
        if isinstance(value, dict):
            clean_dict[key] = sanitize_config(value)
            
        # Handle lists and tuples
        elif isinstance(value, (list, tuple)):
            clean_dict[key] = [sanitize_config(v) for v in value]
            
        # Handle all other types
        else:
            clean_dict[key] = serialize_config(value)
            
    return clean_dict


import os
import json
import datetime
import tempfile
import warnings
import torch
import numpy as np
from rdkit import Chem
from ..utils.checker import MolecularInputChecker
from ..utils.checkpoint import LocalCheckpointManager, HuggingFaceCheckpointManager
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from typing import Optional, ClassVar, Union, List, Dict, Any, Tuple, Callable, Type, Literal

@dataclass
class BaseMolecularEncoder(ABC):
    """Base class for molecular representation learning.

    This class provides a skeleton for implementing custom molecular representation
    learning models. It includes basic functionality for parameter management, 
    fitting, and encoding molecular representations.

    Attributes
    ----------
    model_name : str
        Name of the model
    device : Optional[torch.device]
        Device to run the model on
    model : Optional[torch.nn.Module]
        Model instance
    """
    
    model_name: str = field(default="BaseMolecularEncoder")
    device: Optional[torch.device] = field(default=None)
    model: Optional[torch.nn.Module] = field(default=None, init=False)
    
    # Instance variables with default values
    is_fitted_: bool = field(default=False, init=False)
    model_class: Optional[Type[torch.nn.Module]] = field(default=None, init=False)
    
    def __post_init__(self):
        """Set device if not provided."""
        # Set device if not provided
        if self.device is None:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        elif isinstance(self.device, str):
            self.device = torch.device(self.device)
    
    def get_params(self, deep: bool = True) -> Dict[str, Any]:
        """Get parameters for this estimator.

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key)
            if deep and hasattr(value, "get_params"):
                deep_items = value.get_params().items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params) -> "BaseMolecularEncoder":
        """Set the parameters of this representer.

        Parameters
        ----------
        **params : dict
            Representer parameters.

        Returns
        -------
        self : object
            Representer instance.
        """
        if not params:
            return self

        valid_params = self.get_params(deep=True)
        for key, value in params.items():
            if key not in valid_params:
                raise ValueError(f"Invalid parameter {key} for representer {self}")
            setattr(self, key, value)
        return self

    def _setup_optimizers(self) -> Tuple[torch.optim.Optimizer, Optional[Any]]:
        raise NotImplementedError("optimizers and schedulers must be implemented by child class")

    @abstractmethod
    def fit(self, X: List[str], y: Optional[np.ndarray] = None) -> "BaseMolecularEncoder":
        pass

    @abstractmethod
    def encode(self, X: List[str], return_type: Literal["np", "pt"] = "pt") -> Union[np.ndarray, torch.Tensor]:
        pass

    def _train_epoch(self, train_loader, optimizer):
        raise NotImplementedError("Training epoch logic must be implemented by child class")

    def _evaluation_epoch(self, evaluate_loader):
        raise NotImplementedError("Evaluation epoch logic must be implemented by child class")

    def _check_is_fitted(self) -> None:
        if not self.is_fitted_:
            raise AttributeError(
                "This MolecularBaseRepresenter instance is not fitted yet. "
                "Call 'fit' before using this representer."
            )

    def __repr__(self, N_CHAR_MAX=700):
        """Get string representation of the representer.

        Parameters
        ----------
        N_CHAR_MAX : int, default=700
            Maximum number of non-blank characters to show

        Returns
        -------
        str
            String representation of the representer
        """
        # Get all relevant attributes (excluding private ones and callables)
        attributes = {
            name: value
            for name, value in sorted(self.__dict__.items())
            if not name.startswith("_") and not callable(value)
        }

        def format_value(v):
            """Format a value for representation."""
            if isinstance(v, (float, np.float32, np.float64)):
                return f"{v:.3g}"
            elif isinstance(v, (list, tuple, np.ndarray)) and len(v) > 6:
                return f"{str(v[:3])[:-1]}, ..., {str(v[-3:])[1:]}"
            elif isinstance(v, str) and len(v) > 50:
                return f"'{v[:25]}...{v[-22:]}'"
            elif isinstance(v, dict) and len(v) > 6:
                items = list(v.items())
                return f"{dict(items[:3])}...{dict(items[-3:])}"
            elif isinstance(v, torch.nn.Module):
                return f"{v.__class__.__name__}(...)"
            elif v is None:
                return "None"
            else:
                return repr(v)

        # Build the representation string
        class_name = self.__class__.__name__
        attributes_str = []

        # First add important attributes that should appear first
        important_attrs = ["model_name", "is_fitted_"]
        for attr in important_attrs:
            if attr in attributes:
                value = attributes.pop(attr)
                attributes_str.append(f"{attr}={format_value(value)}")

        # Then add the rest
        sorted_attrs = sorted(attributes.items())
        attributes_str.extend(f"{name}={format_value(value)}" for name, value in sorted_attrs)

        # Join all parts
        content = ",\n    ".join(attributes_str)
        repr_ = f"{class_name}(\n    {content}\n)"

        # Handle length limit
        if len(repr_) > N_CHAR_MAX:
            lines = repr_.split("\n")

            # Keep header
            result = [lines[0]]
            current_length = len(lines[0])

            # Process middle lines
            for line in lines[1:-1]:
                if current_length + len(line) + 5 > N_CHAR_MAX:  # 5 for "...\n"
                    result.append("    ...")
                    break
                result.append(line)
                current_length += len(line) + 1  # +1 for newline

            # Add closing parenthesis
            result.append(lines[-1])
            repr_ = "\n".join(result)

        return repr_

    @abstractmethod
    def _get_model_params(self, checkpoint: Optional[Dict] = None) -> Dict[str, Any]:
        raise NotImplementedError(
            "Method _get_model_params must be implemented by child class. "
            "This method should return a dictionary of model parameters either "
            "from a checkpoint or current instance values."
        )

    def _initialize_model(
        self, model_class: Type[torch.nn.Module], device, checkpoint: Optional[Dict] = None
    ) -> None:
        """Initialize the model with given parameters or checkpoint.
        """
        try:
            model_params = self._get_model_params(checkpoint)

            # Initialize model with parameters
            self.model = model_class(**model_params)

            # Move model to device
            try:
                self.model = self.model.to(device)
            except Exception as e:
                raise RuntimeError(f"Failed to move model to device {device}: {str(e)}")

            # Load state dict if provided in checkpoint
            if checkpoint is not None:
                try:
                    self.model.load_state_dict(checkpoint["model_state_dict"])
                except Exception as e:
                    raise ValueError(f"Failed to load model state dictionary: {str(e)}")

        except Exception as e:
            raise RuntimeError(f"Model initialization failed: {str(e)}")

    @staticmethod
    def _get_param_names() -> List[str]:
        return ["model_name", "is_fitted_"]
    
    def _validate_inputs(
        self, X: List[str], y: Optional[Union[List, np.ndarray]] = None, return_rdkit_mol: bool = True, num_task: int = 0, predefined_num_task: int = 0
    ) -> Tuple[Union[List[str], List["Mol"]], Optional[np.ndarray]]:
        return MolecularInputChecker.validate_inputs(X, y, num_task, predefined_num_task, return_rdkit_mol)

    def _inspect_task_types(self, y: Union[np.ndarray, torch.Tensor], return_type: Literal["pt", "np"] = "pt") -> Union[np.ndarray, torch.Tensor]:
        """Inspect the task types of the target values.

        Parameters
        ----------
        y : Union[np.ndarray, torch.Tensor]
            Target values: 2D array for multiple tasks
        return_type : Literal["pt", "np"], default="pt"
            Return type of the result

        Returns
        -------
        result : Union[np.ndarray, torch.Tensor]
            Result of the task types inspection
        """
        if isinstance(y, np.ndarray):
            y = torch.from_numpy(y)
        # 0/1/nan -> binary classification (True), otherwise -> regression (False)
        result = torch.tensor([torch.all(torch.isnan(y[:, i]) | (y[:, i] == 0) | (y[:, i] == 1)) for i in range(y.shape[1])], dtype=torch.bool)
        if return_type == "np":
            return result.numpy()
        return result

    def save_to_local(self, path: str) -> None:
        """Save to local disk using LocalCheckpointManager."""
        LocalCheckpointManager.save_model_to_local(self, path)

    def load_from_local(self, path: str) -> None:
        """Load from local disk using LocalCheckpointManager."""
        LocalCheckpointManager.load_model_from_local(self, path)

    def save_to_hf(
        self,
        repo_id: str,
        task_id: str = "default",
        metadata_dict: Optional[Dict[str, Any]] = None,
        metrics: Optional[Dict[str, float]] = None,
        commit_message: str = "Update model",
        token: Optional[str] = None,
        private: bool = False,
    ) -> None:
        HuggingFaceCheckpointManager.push_to_huggingface(
            model_instance=self,
            repo_id=repo_id,
            task_id=task_id,
            metadata_dict=metadata_dict,
            metrics=metrics,
            commit_message=commit_message,
            token=token,
            private=private,
        )

    def load_from_hf(self, repo_id: str, path: str) -> None:
        """Download and load model from Hugging Face Hub."""
        HuggingFaceCheckpointManager.load_model_from_hf(self, repo_id, path)

    def save(self, path: Optional[str] = None, repo_id: Optional[str] = None, **kwargs) -> None:
        """
        Automatic save. If `repo_id` is given, push to Hugging Face Hub.
        Otherwise, save to local path.
        """
        if repo_id is not None:
            self.save_to_hf(repo_id=repo_id, **kwargs)
        else:
            if path is None:
                raise ValueError("path must be provided if repo_id is not given.")
            self.save_to_local(path)

    def load(self, path: str, repo_id: Optional[str] = None) -> None:
        """
        Automatic load. If local `path` exists, load from there. 
        Otherwise, try loading from Hugging Face Hub if `repo_id` is given.
        """
        if os.path.exists(path):
            self.load_from_local(path)
        else:
            if repo_id is None:
                raise FileNotFoundError(
                    f"No local file found at '{path}' and no repo_id provided."
                )
            self.load_from_hf(repo_id, path)


import torch

def init_weights(net, init_type='xavier', init_gain=0.02, verbose=False):
    """Initialize network weights.
    Parameters:
        net (network)   -- network to be initialized
        init_type (str) -- the name of an initialization method: normal | xavier | kaiming | orthogonal
        init_gain (float)    -- scaling factor for normal, xavier and orthogonal.
    """
    def init_func(m):  # define the initialization function
        classname = m.__class__.__name__
        if hasattr(m, 'weight') and (classname.find('Conv') != -1 or classname.find('Linear') != -1):
            if init_type == 'normal':
                torch.nn.init.normal_(m.weight.data, 0.0, init_gain)
            elif init_type == 'xavier':
                torch.nn.init.xavier_normal_(m.weight.data, gain=init_gain)
            elif init_type == 'kaiming':
                torch.nn.init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
            elif init_type == 'orthogonal':
                torch.nn.init.orthogonal_(m.weight.data)
            else:
                raise NotImplementedError('initialization method [%s] is not implemented' % init_type)
            if hasattr(m, 'bias') and m.bias is not None:
                torch.nn.init.constant_(m.bias.data, 0.0)
        elif classname.find('BatchNorm2d') != -1:  # BatchNorm Layer's weight is not a matrix
            torch.nn.init.normal_(m.weight.data, 1.0, init_gain)
            torch.nn.init.constant_(m.bias.data, 0.0)
    if verbose:
        print('initialize network with %s' % init_type)
    net.apply(init_func)  # apply the initialization function


from tqdm import tqdm
from dataclasses import dataclass, field
from typing import Optional, Dict, Any, Tuple, List, Type
import numpy as np

import torch

from .jtnn_vae import JTNNVAE
from .jtnn.mol_tree import MolTree
from .jtnn.vocab import Vocab
from .jtnn.datautils import MolTreeFolder
from ...base import BaseMolecularGenerator

@dataclass
class JTVAEMolecularGenerator(BaseMolecularGenerator):
    """ 
    JT-VAE-based molecular generator. Implemented for unconditional moleculargeneration.     

    References
    ----------
    - Junction Tree Variational Autoencoder for Molecular Graph Generation. ICML 2018. https://arxiv.org/pdf/1802.04364
    - Code: https://github.com/kamikaze0923/jtvae

    Parameters
    ----------
    hidden_size : int, default=450
        Dimension of hidden layers in the model.
    latent_size : int, default=56
        Dimension of the latent space.
    depthT : int, default=20
        Depth of the tree encoder.
    depthG : int, default=3
        Depth of the graph decoder.
    batch_size : int, default=32
        Number of samples per batch during training.
    epochs : int, default=20
        Number of epochs to train the model.
    learning_rate : float, default=0.003
        Initial learning rate for the optimizer.
    weight_decay : float, default=0.0
        L2 regularization factor.
    grad_norm_clip : Optional[float], default=None
        Maximum norm for gradient clipping. None means no clipping.
    beta : float, default=0.0
        Initial KL divergence weight for VAE training.
    step_beta : float, default=0.002
        Step size for KL annealing.
    max_beta : float, default=1.0
        Maximum value for KL weight.
    warmup : int, default=40000
        Number of steps for KL annealing warmup.
    use_lr_scheduler : bool, default=True
        Whether to use learning rate scheduling.
    anneal_rate : float, default=0.9
        Learning rate annealing factor.
    anneal_iter : int, default=40000
        Number of iterations between learning rate updates.
    kl_anneal_iter : int, default=2000
        Number of iterations between KL weight updates.
    verbose : bool, default=False
        Whether to print detailed training information.
    """
    # Model parameters
    hidden_size: int = 450
    latent_size: int = 56
    depthT: int = 20
    depthG: int = 3

    # Training parameters
    batch_size: int = 32
    epochs: int = 20
    learning_rate: float = 0.003
    weight_decay: float = 0.0
    grad_norm_clip: Optional[float] = None
    
    # KL annealing parameters
    beta: float = 0.0
    step_beta: float = 0.002
    max_beta: float = 1.0
    warmup: int = 40000

    # Learning rate scheduling
    use_lr_scheduler: bool = True
    anneal_rate: float = 0.9
    anneal_iter: int = 40000
    kl_anneal_iter: int = 2000

    # Other parameters
    verbose: bool = False

    # attributes
    model_name: str = "JTVAEMolecularGenerator"
    fitting_loss: List[float] = field(default_factory=list, init=False)
    fitting_epoch: int = field(default=0, init=False)
    model_class: Type[JTNNVAE] = field(default=JTNNVAE, init=False)
    
    def __post_init__(self):
        super().__post_init__()
        self.vocab = None

    @staticmethod
    def _get_param_names() -> List[str]:
        return [
            "hidden_size", "latent_size", "depthT", "depthG",
            "batch_size", "epochs", "learning_rate", "weight_decay",
            "beta", "step_beta", "max_beta", "warmup",
            "use_lr_scheduler", "anneal_rate", "anneal_iter", "kl_anneal_iter",
            "verbose", 'model_name', 'vocab'
        ]
    
    def _get_model_params(self, checkpoint: Optional[Dict] = None) -> Dict[str, Any]:
        params = [
            "vocab", "hidden_size", "latent_size", "depthT", "depthG"
        ]
        if checkpoint is not None:
            if "hyperparameters" not in checkpoint:
                raise ValueError("Checkpoint missing 'hyperparameters' key")
            return {k: checkpoint["hyperparameters"][k] for k in params}
    
        return {k: getattr(self, k) for k in params}

    def _setup_optimizers(self) -> Tuple[torch.optim.Optimizer, Optional[Any]]:
        """Setup optimization components including optimizer and learning rate scheduler.
        """
        optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay
        )
        scheduler = None
        if self.use_lr_scheduler:
            scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, self.anneal_rate)

        return optimizer, scheduler

    def _extract_vocab(self, X):
        cset = set()
        for ii, smiles in enumerate(X):
            try:
                mol = MolTree(smiles)
                for c in mol.nodes:
                    cset.add(c.smiles)
            except Exception as e:
                print(f'Error {e} in extracting vocab for smiles: {smiles}')
                pass
        vocab = list(cset)
        return vocab

    def _convert_to_tensor(self, X):
        all_data = []
        for smiles in X:
            try:
                mol_tree = MolTree(smiles)
                mol_tree.recover()
                mol_tree.assemble()
                for node in mol_tree.nodes:
                    if node.label not in node.cands:
                        node.cands.append(node.label)

                del mol_tree.mol
                for node in mol_tree.nodes:
                    del node.mol
            except Exception as e:
                print(f'Error {e} in tensorizing smiles: {smiles}')
                mol_tree = None
            all_data.append(mol_tree)
        return all_data

    def fit(
        self,
        X_train: List[str],
    ) -> "JTVAEMolecularGenerator":

        X_train, _ = self._validate_inputs(X_train, None, num_task=0, return_rdkit_mol=False)
        vocab = self._extract_vocab(X_train)
        vocab = Vocab(vocab)
        self.vocab = vocab

        self._initialize_model(self.model_class)
        self.model.initialize_parameters()
        optimizer, scheduler = self._setup_optimizers()

        train_dataset = self._convert_to_tensor(X_train)
        train_dataset = list(filter(lambda x: x is not None, train_dataset))
        train_loader = MolTreeFolder(train_dataset, vocab, self.batch_size, num_workers=0)
        step_len = len(X_train) // self.batch_size
        self.fitting_loss = []
        self.fitting_epoch = 0
        total_step = 0
        for epoch in range(self.epochs):
            train_losses, total_step = self._train_epoch(train_loader, optimizer, scheduler, epoch, total_step, step_len)
            self.fitting_loss.append(np.mean(train_losses).item())

        self.fitting_epoch = epoch
        self.is_fitted_ = True
        return self

    def _train_epoch(self, train_loader, optimizer, scheduler, epoch, total_step, step_len):
        self.model.train()
        losses = []
        iterator = (
            tqdm(train_loader, desc="Training", leave=False, total=step_len)
            if self.verbose
            else train_loader
        )
        for step, batched_data in enumerate(iterator):
            total_step += 1
            optimizer.zero_grad()

            word_loss, topo_loss, assm_loss, kl_div = self.model.compute_loss(batched_data)
            total_loss = word_loss + topo_loss + assm_loss + kl_div * self.beta
            total_loss.backward()
            if self.grad_norm_clip is not None:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_norm_clip)
            optimizer.step()
            losses.append(total_loss.item())

            if total_step % self.anneal_iter == 0:
                scheduler.step()
                
            if total_step % self.kl_anneal_iter == 0 and total_step >= self.warmup:
                self.beta = min(self.max_beta, self.beta + self.step_beta)

            if self.verbose:
                iterator.set_postfix({"Epoch": epoch, "Loss": f"{total_loss.item():.4f}, word_loss: {word_loss.item():.4f}, topo_loss: {topo_loss.item():.4f}, assm_loss: {assm_loss.item():.4f}, kl_div: {kl_div.item():.4f}"})
            
        return losses, total_step

    def generate(
        self, 
        batch_size: int = 32
    ) -> List[str]:
        """Generate molecules using JT-VAE.
        
        Parameters
        ----------  
        batch_size : int
            Number of molecules to generate.
            
        Returns
        -------
        List[str]
            Generated molecules as SMILES strings.
        """
        if not self.is_fitted_:
            raise ValueError("Model must be fitted before generating molecules.")
        
        return self.model.sample_prior(self.device)
    



import warnings
import torch
import numpy as np
from ..utils import (
    roc_auc_score,
    accuracy_score,
    mean_absolute_error,
    mean_squared_error,
    root_mean_squared_error,
    r2_score,
)
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from typing import Optional, ClassVar, Union, List, Dict, Any, Tuple, Callable, Type
from ..base.base import BaseModel

@dataclass
class BaseMolecularPredictor(BaseModel, ABC):
    """Base class for molecular discovery estimators."""
    
    model_name: str = field(default="BaseMolecularPredictor")
    num_task: int = field(default=0)
    task_type: str = field(default=None)
    DEFAULT_METRICS: ClassVar[Dict] = {
        "classification": {"default": ("roc_auc", roc_auc_score, True)},
        "regression": {"default": ("mae", mean_absolute_error, False)},
    }

    def __post_init__(self):
        super().__post_init__()
        if self.task_type not in ["classification", "regression"]:
            raise ValueError(f"Invalid task_type: {self.task_type}")
        if self.num_task <= 0:
            raise ValueError(f"num_task must be positive, got {self.num_task}")

    @staticmethod
    def _get_param_names(self) -> List[str]:
        return super()._get_param_names() + ["num_task", "task_type"]

    @abstractmethod
    def autofit(self, X_train, y_train, X_val=None, y_val=None, search_parameters: Optional[dict] = None, n_trials: int = 10) -> "BaseMolecularPredictor": 
        pass
    
    @abstractmethod
    def fit(self, X_train, y_train, X_val=None, y_val=None, search_parameters: Optional[dict] = None, n_trials: int = 10) -> "BaseMolecularPredictor": 
        pass

    @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def _evaluation_epoch(self, evaluate_loader):
        pass
        
    def _setup_evaluation(
        self,
        evaluate_criterion: Optional[Union[str, Callable]],
        evaluate_higher_better: Optional[bool],
    ) -> None:
        if evaluate_criterion is None:
            default_metric = self.DEFAULT_METRICS[self.task_type]["default"]
            self.evaluate_name = default_metric[0]
            self.evaluate_criterion = default_metric[1]
            self.evaluate_higher_better = default_metric[2]
        else:
            if isinstance(evaluate_criterion, str):
                metric_map = {
                    "accuracy": (accuracy_score, True),
                    "roc_auc": (roc_auc_score, True),
                    "rmse": (root_mean_squared_error, False),
                    "mse": (mean_squared_error, False),
                    "mae": (mean_absolute_error, False),
                    "r2": (r2_score, True),
                }
                if evaluate_criterion not in metric_map:
                    raise ValueError(
                        f"Unknown metric: {evaluate_criterion}. "
                        f"Available metrics: {list(metric_map.keys())}"
                    )
                self.evaluate_name = evaluate_criterion
                self.evaluate_criterion = metric_map[evaluate_criterion][0]
                self.evaluate_higher_better = (
                    metric_map[evaluate_criterion][1]
                    if evaluate_higher_better is None
                    else evaluate_higher_better
                )
            else:
                if evaluate_higher_better is None:
                    raise ValueError(
                        "evaluate_higher_better must be specified for a custom function."
                    )
                self.evaluate_name = "custom"
                self.evaluate_criterion = evaluate_criterion
                self.evaluate_higher_better = evaluate_higher_better
    
    def _load_default_criterion(self):
        if self.task_type == "regression":
            return torch.nn.L1Loss(reduction='none')
        elif self.task_type == "classification":
            return torch.nn.BCEWithLogitsLoss(reduction='none')
        else:
            warnings.warn(
                "Unknown task type. Using L1 Loss as default. "
                "Please specify 'regression' or 'classification' for better results."
            )
            return torch.nn.L1Loss(reduction='none')
    
    def _validate_inputs(
        self, X: List[str], y: Optional[Union[List, np.ndarray]] = None, num_task: int = 0, num_pretask: int = 0, return_rdkit_mol: bool = True
    ) -> Tuple[Union[List[str], List["Chem.Mol"]], Optional[np.ndarray]]:
        return super()._validate_inputs(X, y, self.num_task, 0, return_rdkit_mol)